from dataclasses import dataclass, field
from typing import Any, ClassVar

import pytest

import cappa
from cappa.command import Command
from tests.utils import Backend, backends, require_dataclass_kw_only


@backends
def test_color_off(backend: Backend):
    def no_op():
        pass

    @cappa.command(invoke=no_op)
    @dataclass
    class Example: ...

    cappa.invoke(Example, color=False, argv=[], backend=backend)


@backends
def test_no_help(backend: Backend):
    @dataclass
    class Example: ...

    result = cappa.parse(Example, argv=[], help=False, backend=backend)
    assert result == Example()

    with pytest.raises(SystemExit) as e:
        cappa.parse(Example, argv=["-h"], help=False, backend=backend)
    assert e.value.code == 2

    with pytest.raises(SystemExit) as e:
        cappa.parse(Example, argv=["--help"], help=False, backend=backend)
    assert e.value.code == 2


@pytest.mark.help
@backends
def test_arg_help(capsys: Any, backend: Backend):
    @dataclass
    class Example: ...

    help: cappa.Arg[str] = cappa.Arg(short="-p", long="--pelp")

    result = cappa.parse(Example, argv=[], help=help, backend=backend)
    assert result == Example()

    with pytest.raises(SystemExit) as e:
        cappa.parse(Example, argv=["-p"], help=help, backend=backend)
    assert e.value.code == 0

    out = capsys.readouterr().out
    assert "-p, --pelp" in out

    with pytest.raises(SystemExit) as e:
        cappa.parse(Example, argv=["--pelp"], help=help, backend=backend)
    assert e.value.code == 0

    out = capsys.readouterr().out
    assert "-p, --pelp" in out


@backends
def test_version_enabled(capsys: Any, backend: Backend):
    @dataclass
    class Example: ...

    with pytest.raises(SystemExit) as e:
        cappa.parse(Example, argv=["--version"], version="1.2.3", backend=backend)
    assert e.value.code == 0

    out = capsys.readouterr().out
    assert "1.2.3" in out


@backends
def test_arg_version(capsys: Any, backend: Backend):
    @dataclass
    class Example: ...

    version: cappa.Arg[str] = cappa.Arg("1.2.3", short="-p", long="--persion")

    result = cappa.parse(Example, argv=[], version=version, backend=backend)
    assert result == Example()

    with pytest.raises(SystemExit) as e:
        cappa.parse(Example, argv=["-p"], version=version, backend=backend)
    assert e.value.code == 0

    out = capsys.readouterr().out
    assert "1.2.3" in out

    with pytest.raises(SystemExit) as e:
        cappa.parse(Example, argv=["--persion"], version=version, backend=backend)
    assert e.value.code == 0

    out = capsys.readouterr().out
    assert "1.2.3" in out


@backends
def test_version_without_help(backend: Backend):
    @dataclass
    class Example: ...

    result = cappa.parse(Example, argv=[], version="1.2.3", help=False, backend=backend)
    assert result == Example()


@backends
def test_arg_explicit_version_missing_name(backend: Backend):
    @dataclass
    class Example: ...

    version: cappa.Arg[str] = cappa.Arg(short="-p", long="--persion")

    with pytest.raises(ValueError) as e:
        cappa.parse(Example, argv=["-p"], version=version, backend=backend)

    assert (
        str(e.value)
        == "Expected explicit version `Arg` to supply version number as its name, like `Arg('1.2.3', ...)`"
    )


@backends
def test_arg_explicit_version_long_true_defaults_to_version(
    capsys: Any, backend: Backend
):
    @dataclass
    class Example: ...

    version: cappa.Arg[str] = cappa.Arg("1.2.3", short="-p", long=True)

    with pytest.raises(cappa.Exit):
        cappa.parse(Example, argv=["--version"], version=version, backend=backend)

    out = capsys.readouterr().out
    assert "1.2.3" in out


@backends
def test_prog_basename(capsys: Any, backend: Backend):
    @dataclass
    class Example: ...

    with pytest.raises(cappa.Exit):
        cappa.parse(Example, argv=["--help"], backend=backend)

    out = capsys.readouterr().out
    assert "usage: example" in out.lower()


@backends
def test_collect_composes_with_parse(backend: Backend):
    @dataclass
    class Example: ...

    command: Command[Example] = cappa.collect(Example, backend=backend)
    result = cappa.parse(command, argv=[], backend=backend)

    assert result == Example()


@require_dataclass_kw_only
@backends
def test_kw_only_base_class_field_ordering(backend: Backend):
    kwargs = {"kw_only": True}

    @dataclass(**kwargs)
    class Base:
        name: str = "hello"

    @dataclass
    class Child(Base):
        number: int = 0

    result = cappa.parse(
        cappa.Command(Child, default_long=True),
        argv=["--name", "world"],
        backend=backend,
    )
    assert result == Child(number=0, name="world")


@backends
def test_collect_skips_non_init_fields(backend: Backend):
    @dataclass
    class Example:
        skipped1: ClassVar[str] = "skipped1"
        skipped2: str = field(default="", init=False)
        present: str = ""

    command: Command[Example] = cappa.collect(Example, backend=backend)
    result = cappa.parse(command, argv=[], backend=backend)

    assert result == Example()



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import parse_completion


def foo(f: str):
    return [cappa.Completion(f"{f}ooooo", "ok")]


def test_custom_completion():
    @dataclass
    class Args:
        value: Annotated[str, cappa.Arg(short=True, completion=foo)]

    result = parse_completion(Args, "-v", "f")
    assert result
    assert result == "fooooo:ok"



from __future__ import annotations

from dataclasses import dataclass, field

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_list_option(backend: Backend):
    @dataclass
    class ArgTest:
        variable_number: Annotated[set[str], cappa.Arg(short=True, long=True)] = field(
            default_factory=lambda: set()
        )

    test = parse(
        ArgTest,
        "-v",
        "one",
        "--variable-number",
        "one",
        "--variable-number",
        "two",
        "-v",
        "two",
        backend=backend,
    )
    assert test.variable_number == {"one", "two"}


@backends
def test_list_positional(backend: Backend):
    @dataclass
    class ArgTest:
        variable_number: set[int] = field(default_factory=lambda: set())

    test = parse(
        ArgTest,
        "1",
        "2",
        "3",
        "2",
        "3",
        "1",
        backend=backend,
    )
    assert test.variable_number == {1, 2, 3}



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

from typing_extensions import Annotated

import cappa
from tests.utils import parse_completion


@dataclass
class Foo: ...


@dataclass
class Bar:
    nested_opt: Annotated[str, cappa.Arg(long=True)]


@dataclass
class Args:
    value: cappa.Subcommands[Union[Foo, Bar]]


def test_subcommand_name_no_partial():
    result = parse_completion(Args, "")
    assert result == "foo:\nbar:"


def test_subcommand_name_with_partial():
    result = parse_completion(Args, "f")
    assert result
    assert result == "foo:"


def test_subcommand_args():
    result = parse_completion(Args, "bar", "--n")
    assert result
    assert result == "--nested-opt:<nested_opt>"



from dataclasses import dataclass

import cappa
from tests.utils import Backend, backends, invoke


@dataclass
class SubSubCommand:
    """Sub-subcommand."""

    def __call__(self) -> int:
        return 4


@dataclass
class SubCommand:
    subcommand: cappa.Subcommands[SubSubCommand]


@dataclass
class MainCommand:
    subcommand: cappa.Subcommands[SubCommand]


@backends
def test_parent_commands_dont_require_invoke_fn(backend: Backend):
    result = invoke(MainCommand, "sub-command", "sub-sub-command", backend=backend)
    assert result == 4



from __future__ import annotations

from dataclasses import dataclass

import pytest

import cappa
from tests.utils import Backend, backends, invoke


@backends
def test_missing_invoke(backend: Backend):
    @cappa.command()
    @dataclass
    class Command: ...

    with pytest.raises(RuntimeError) as e:
        invoke(Command, backend=backend)

    exc = e.value
    cause = exc.__cause__
    assert "due to resolution failure" in str(exc)
    assert (
        f"Cannot call `invoke` for a command which does not have an invoke handler: {Command}."
        == str(cause)
    )


@backends
def test_string_missing_function(backend: Backend):
    @cappa.command(invoke="sys")
    @dataclass
    class Command: ...

    with pytest.raises(RuntimeError) as e:
        invoke(Command, backend=backend)

    exc = e.value
    cause = exc.__cause__
    assert "due to resolution failure" in str(exc)
    assert (
        "Invoke `sys` must be a fully qualified reference to a function in a module."
        == str(cause)
    )


@backends
def test_string_references_invalid_module(backend: Backend):
    @cappa.command(invoke="completely.wrong")
    @dataclass
    class Command: ...

    with pytest.raises(RuntimeError) as e:
        invoke(Command, backend=backend)

    exc = e.value
    cause = exc.__cause__
    assert "due to resolution failure" in str(exc)
    assert "No module 'completely' when attempting to load 'completely.wrong'." == str(
        cause
    )


@backends
def test_string_references_invalid_function(backend: Backend):
    @cappa.command(invoke="builtins.meow")
    @dataclass
    class Command: ...

    with pytest.raises(RuntimeError) as e:
        invoke(Command, backend=backend)

    exc = e.value
    cause = exc.__cause__
    assert "due to resolution failure" in str(exc)
    assert (
        "Module <module 'builtins' (built-in)> does not have a function `meow`."
        == str(cause)
    )


@backends
def test_string_reference_not_callable(backend: Backend):
    @cappa.command(invoke="builtins.__name__")
    @dataclass
    class Command: ...

    with pytest.raises(RuntimeError) as e:
        invoke(Command, backend=backend)

    exc = e.value
    cause = exc.__cause__
    assert "due to resolution failure" in str(exc)
    assert "`builtins` does not reference a valid callable." == str(cause)



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

import pytest

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_unrecognized_post_dash_arg(backend: Backend):
    @dataclass
    class Args:
        foo: str
        raw: Union[str, None] = None

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "foovalue", "raw", "raw2", backend=backend)

    assert e.value.code == 2
    assert "unrecognized arguments: raw2" in str(e.value.message).lower()



from __future__ import annotations

from dataclasses import dataclass

import cappa
from tests.utils import parse


@dataclass
class Args:
    foo: int = 4
    bar: int = 5


@cappa.command(default_long=True, default_short=True)
@dataclass
class Command:
    args: cappa.Destructured[Args]


@cappa.command(default_long=True)
@dataclass
class InnerOverrideArgs:
    foo: int = 4


@cappa.command(default_long=False)
@dataclass
class InnerOverrideCommand:
    args: cappa.Destructured[InnerOverrideArgs]


def test_default_long_propagates():
    test = parse(Command, "--foo", "1", "--bar", "2")
    assert test == Command(Args(1, 2))


def test_default_short_propagates():
    test = parse(Command, "-f", "1", "-b", "2")
    assert test == Command(Args(1, 2))


def test_inner_command_setting_takes_precedence():
    test = parse(InnerOverrideCommand, "--foo", "1")
    assert test == InnerOverrideCommand(InnerOverrideArgs(1))



from __future__ import annotations

import contextlib
from dataclasses import dataclass
from typing import Any

import cappa
from tests.utils import Backend, backends


@dataclass
class Command:
    value: str = "test"
    closed: bool = False

    def __enter__(self):
        return self

    def __exit__(self, *args: Any):
        object.__setattr__(self, "closed", True)


@backends
def test(backend: Backend):
    with contextlib.ExitStack() as stack:
        result = cappa.parse(Command, argv=[], backend=backend, exit_stack=stack)

        assert not result.closed
        assert result.value == "test"

    assert result.closed



from __future__ import annotations

from dataclasses import dataclass

import pytest

import cappa
from tests.utils import parse


@dataclass
class Foo:
    foo: int


@dataclass
class CLI:
    attrs: cappa.Destructured[Foo | int]


def test_destructured_union():
    with pytest.raises(ValueError) as e:
        parse(CLI)
    assert (
        str(e.value)
        == "Destructured arguments currently only support singular concrete types."
    )



from sphinx_pyproject import SphinxConfig

config = SphinxConfig("../../pyproject.toml")

project = "Cappa"
copyright = "2023, Dan Cardin"
author = config.author
release = config.version

extensions = [
    "autoapi.extension",
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "sphinx_togglebutton",
    "sphinx_toolbox.collapse",
]

templates_path = ["_templates"]

exclude_patterns = []

html_theme = "furo"
html_theme_options = {
    "navigation_with_keys": True,
}
html_static_path = ["_static"]

myst_heading_anchors = 4
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

autoapi_type = "python"
autoapi_dirs = ["../../src/cappa"]
autoapi_generate_api_docs = False

autodoc_typehints = "description"
autodoc_typehints_format = "fully-qualified"
suppress_warnings = ["autoapi.python_import_resolution"]

autosectionlabel_prefix_document = True



from __future__ import annotations

import asyncio
from dataclasses import dataclass

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends


async def async_parse_that_fails(value: str) -> int:
    await asyncio.sleep(0.001)
    raise ValueError("Async parsing failed")


@dataclass
class Command:
    value: Annotated[int, cappa.Arg(parse=async_parse_that_fails)]

    def __call__(self):
        return self


@backends
def test(backend: Backend):
    with pytest.raises(cappa.Exit) as exc_info:
        asyncio.run(cappa.invoke_async(Command, argv=["42"], backend=backend))

    assert exc_info.value.code == 2
    assert "Async parsing failed" in str(exc_info.value.message)



from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@cappa.command(name="insiders")
@dataclass
class Args:
    arg: Annotated[int, cappa.Arg(short="-ab")]
    foo: Annotated[bool, cappa.Arg(short="-bc")] = False


@backends
def test_single_opt(backend: Backend):
    args = parse(Args, "-ab", "1", backend=backend)
    assert args == Args(arg=1, foo=False)



from __future__ import annotations

import re
from dataclasses import dataclass
from textwrap import dedent
from typing import Any

import pytest

import cappa
from cappa.output import Exit
from tests.utils import (
    Backend,
    backends,
    parse,
    strip_trailing_whitespace,
    terminal_width,
)


@pytest.mark.help
@backends
def test_default_help(backend: Backend, capsys: Any):
    @dataclass
    class Command:
        """Some help.

        More detail.
        """

        ...

    with pytest.raises(Exit):
        parse(Command, "--help", backend=backend)

    stdout = capsys.readouterr().out
    assert re.match(r".*Some help\.\s+More detail\..*", stdout, re.DOTALL)


@pytest.mark.help
@backends
def test_default_help_no_long_description(backend: Backend, capsys: Any):
    @dataclass
    class Command:
        """Some help."""

        ...

    with pytest.raises(Exit):
        parse(Command, "--help", backend=backend)

    stdout = capsys.readouterr().out
    assert "Some help." in stdout


@pytest.mark.help
@backends
def test_unannotated_argument(backend: Backend, capsys: Any):
    @cappa.command(help="All the help.")
    @dataclass
    class Command: ...

    with pytest.raises(Exit):
        parse(Command, "--help", backend=backend)

    stdout = capsys.readouterr().out
    assert "All the help." in stdout


@pytest.mark.help
@backends
def test_description_without_help(backend: Backend, capsys: Any):
    @cappa.command(description="All the help.")
    @dataclass
    class Command:
        pass

    with pytest.raises(Exit):
        parse(Command, "--help", backend=backend)

    stdout = capsys.readouterr().out
    assert "All the help." in stdout


@pytest.mark.help
@backends
def test_epilog(backend: Backend, capsys: Any):
    @cappa.command(epilog="See also: https://example.com")
    @dataclass
    class Command:
        pass

    with terminal_width(), pytest.raises(Exit):
        parse(Command, "--help", backend=backend, completion=False)

    result = strip_trailing_whitespace(capsys.readouterr().out)
    tail = result[result.index("[-h, --help]") :]
    assert tail == dedent(
        """\
        [-h, --help]  Show this message and exit.


          See also: https://example.com
        """
    )



"""Define the external facing argument definition types provided by a user."""

from __future__ import annotations

import dataclasses
import enum
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    Literal,
    Sequence,
    Set,
    TextIO,
    Union,
    cast,
)

from typing_extensions import TypeAlias

from cappa.class_inspect import Field, extract_dataclass_metadata
from cappa.completion.completers import complete_choices
from cappa.completion.types import Completion
from cappa.default import Default, DefaultFormatter, ValueFrom
from cappa.invoke.types import Resolved
from cappa.parse import (
    Parser,
    evaluate_parse,
    parse_handler,
    parse_literal,
    parse_value,
)
from cappa.state import State
from cappa.type_view import Empty, EmptyType, TypeView
from cappa.typing import (
    Doc,
    DocType,
    T,
    detect_choices,
    find_annotations,
)

if TYPE_CHECKING:
    from cappa.destructure import Destructure


@enum.unique
class ArgAction(enum.Enum):
    """`Arg` action typee.

    Options:
      - set: Stores the given CLI value directly.
      - store_true: Stores a literal `True` value, causing options to not attempt to
        consume additional CLI arguments
      - store_false: Stores a literal `False` value, causing options to not attempt to
        consume additional CLI arguments
      - append: Produces a list, and accumulates the given value on top of prior values.
      - count: Increments an integer starting at 0
      - help: Cancels argument parsing and prints the help text
      - version: Cancels argument parsing and prints the CLI version
      - completion: Cancels argument parsing and enters "completion mode"
    """

    set = "store"
    store_true = "store_true"
    store_false = "store_false"
    append = "append"
    count = "count"

    help = "help"
    version = "version"
    completion = "completion"

    @classmethod
    def meta_actions(cls) -> Set[ArgAction]:
        return {cls.help, cls.version, cls.completion}

    @classmethod
    def is_custom(cls, action: ArgAction | Callable[..., Any] | None):
        return action is not None and not isinstance(action, ArgAction)

    @classmethod
    def is_non_value_consuming(cls, action: ArgAction | Callable[..., Any] | None):
        return action in {
            ArgAction.store_true,
            ArgAction.store_false,
            ArgAction.count,
            ArgAction.version,
            ArgAction.help,
        }

    @property
    def is_bool_action(self):
        return self in {self.store_true, self.store_false}


ArgActionType: TypeAlias = Union[ArgAction, Callable[..., Any]]


@dataclasses.dataclass(frozen=True)
class NumArgs:
    """Control the number of values consumed for an argument, with optional-value support.

    Note: The vast majority of uses of `Arg.num_args` should be covered by supplying an integer value,
    or `NumArgs.unbounded()`.

    When `required=False`, the argument can be supplied without a value (e.g. `--foo`
    with no following value), yielding the `NumArgs.default` rather than the Arg's overall default.
    This maps to argparse's `nargs='?'` / `const` behaviour or click' s `is_flag=True` / `flag_value` behavior.

    Args:
        n: Number of values to consume. ``-1`` means unbounded. Defaults to ``1``.
        required: Whether a value is required when the option flag is present.
            When ``False``, ``--foo`` (no value) yields ``default`` and ``--foo VAL``
            yields the parsed ``VAL``. Defaults to ``True``.
        default: Value used when the flag is present but no value follows.
            Only meaningful when ``required=False``. Defaults to ``None``.

    Example::

        @dataclass
        class Args:
            # test.py          -> Args(foo=5)    (class default, flag not given)
            # test.py --foo    -> Args(foo=None) (flag given, no value → NumArgs.default)
            # test.py --foo 4  -> Args(foo=4)    (flag given with value)
            foo: Annotated[
                int | None, Arg(long=True, num_args=NumArgs(required=False))
            ] = 5
    """

    n: int = 1
    required: bool = True
    default: Any = None

    @classmethod
    def unbounded(cls):
        return cls(n=-1)

    @classmethod
    def infer(
        cls,
        type_view: TypeView[Any],
        field_name: str,
        arg: Arg[Any] | None = None,
        action: ArgActionType | None = None,
        long: list[str] | Literal[False] = False,
    ) -> NumArgs:
        if arg:
            if arg.num_args is not None:
                if isinstance(arg.num_args, NumArgs):
                    return arg.num_args
                return cls(n=arg.num_args)

            if arg.parse:
                return cls()

            if ArgAction.is_non_value_consuming(action):
                return cls(n=0)

        if type_view.is_union:
            # Recursively determine the `n` for each variant.  Accept the
            # result only if all non-None variants agree on the same count.
            distinct_n: set[int] = set()
            n_variants: list[tuple[Any, int]] = []
            for type_arg in type_view.inner_types:
                if type_arg.is_none_type:
                    continue
                inferred = cls.infer(
                    type_arg, field_name, arg=arg, action=action, long=long
                )
                distinct_n.add(inferred.n)
                n_variants.append((type_arg.raw, inferred.n))

            if len(distinct_n) == 1:
                return cls(n=distinct_n.pop())

            invalid_kinds = ", ".join(
                [f"`{t}` produces `num_args={n}`" for t, n in n_variants]
            )
            raise ValueError(
                f"On field '{field_name}', mismatch of arity between union variants. {invalid_kinds}."
            )

        if type_view.is_tuple and not type_view.is_variadic_tuple:
            return cls(n=len(type_view.args))

        is_positional = arg is None or (not arg.short and not long)
        is_sequence = type_view.is_variadic_tuple or type_view.is_subclass_of(
            (
                list,
                set,
            )
        )
        if is_positional and is_sequence:
            return cls.unbounded()

        # Options whose outer type is a sequence derive count from the inner type.
        if type_view.is_subclass_of((list, set, tuple)):
            inner = cls.infer(type_view.inner_types[0], field_name)
            return cls(n=inner.n)

        return cls()

    @staticmethod
    def raw_n(num_args: int | NumArgs | None) -> int | None:
        """Extract the integer count from a raw (pre-normalization) num_args value."""
        if isinstance(num_args, NumArgs):
            return num_args.n
        return num_args

    def __str__(self) -> str:
        return str(self.n)


@dataclasses.dataclass(frozen=True)
class Group:
    """Object used to control argument/subcommand grouping in generated help text.

    Note, when implicit group syntax is used, exclusive groups are automatically inferred,
    and the group identity is inferred to be the argument's `field_name`. E.g.
    `foo: Annotated[bool, Arg(long="--foo"), Arg(long="--bar")]`.

    Args:
        order: A number representing the relative ordering among different argument
            groups. Groups with the same order will be displayed alphabetically by
            name.
        name: The display name of the group in help text.
        exclusive: Used to control whether items in the group are mutually exclusive with each other.
            If `True`, the name of the group will be used to control group identity. If a string
            is provided, items with the same exclusive string will be considered part of the same
            exclusivity group.
        section: A secondary level of ordering. Sections have a higher order precedence
            than ``order``, in order to facilitate meta-ordering amongst kinds of groups
            (such as "meta" arguments (``--help``, ``--version``, etc) and subcommands).
            The default ``section`` for any normal argument/``Group`` is 0, for
            ``Subcommand`` it is 1, and for "meta" arguments it is 2.

    Examples:
        A non-exclusive group which just affects helptext display grouping
        >>> arg = Arg(group=Group(name="Colors"))

        An exclusive group, in which all "Colors"-named arguments are mutually exclusive
        >>> arg = Arg(group=Group(name="Colors", exclusive=True))

        An exclusive group, in which all "colors"-identity arguments are mutually exclusive,
        but whose helptext is grouped alongside all other "Options".
        >>> arg = Arg(group=Group(name="Options", exclusive="colors"))
    """

    order: int = dataclasses.field(default=0, compare=False)
    name: str = ""
    exclusive: bool | str | None = dataclasses.field(default=None, compare=True)
    section: int = 0
    id: str = ""

    @cached_property
    def key(self):
        return (self.section, self.order, self.name, bool(self.exclusive))

    def diff(self, other: Group) -> str:
        """Return a string showing only the attributes that differ between two Groups.

        Returns a string in the format:
        "Group(attr=value, ...) != Group(attr=other_value, ...)"
        """
        differing_attrs: list[str] = ["id"]
        for field in dataclasses.fields(self):
            if not field.compare and field.name != "id":
                continue

            self_value = getattr(self, field.name)
            other_value = getattr(other, field.name)
            if self_value != other_value:
                differing_attrs.append(field.name)

        self_parts = [f"{attr}={getattr(self, attr)!r}" for attr in differing_attrs]
        other_parts = [f"{attr}={getattr(other, attr)!r}" for attr in differing_attrs]

        return f"Group({', '.join(self_parts)}) != Group({', '.join(other_parts)})"

    @classmethod
    def infer(
        cls,
        arg: Arg[Any],
        is_option: bool,
        exclusive: bool | str = False,
    ) -> Group:
        order = 0
        name = None
        section = 0

        if isinstance(arg.group, Group):
            order = arg.group.order
            section = arg.group.section

            if arg.group.name:
                name = arg.group.name
            if arg.group.exclusive is not None:
                exclusive = arg.group.exclusive

        if isinstance(arg.group, str):
            name = arg.group

        if isinstance(arg.group, tuple):
            order, name = arg.group

        if name is None:
            if is_option:
                name = "Options"
            else:
                name = "Arguments"
                order = 1

        if isinstance(exclusive, str):
            identity = exclusive
        else:
            identity = name

        return cls(
            name=name,
            order=order,
            exclusive=bool(exclusive),
            section=section,
            id=identity,
        )


@dataclasses.dataclass(frozen=True)
class Arg(Generic[T]):
    """Describe a CLI argument.

    Arguments:
        value_name: Placeholder for the argument's value in the help message / usage.
            Defaults to the name of the corresponding class field.
        short: If `True`, uses first letter of the name to infer a (ex. `-s`) short
            flag. If a string is supplied, that will be used instead. If a string is supplied,
            it is split on '/' (forward slash), to support multiple options. Additionally
            accepts a list of strings.
        long: If `True`, uses first letter of the name to infer a (ex. `--long`) long
            flag. If a string is supplied, that will be used instead. If a string is supplied,
            it is split on '/' (forward slash), to support multiple options. Additionally
            accepts a list of strings.
        count: If `True` the resultant argument will count instances and accept zero
            arguments.
        default: An explicit default CLI value. When left unspecified, the default is
            inferred from the class' default or the adapter default/default_factory.
            Note, if the default value is used, it will not be coerced into the annotated
            type through the `parse` method.
        help: By default, the help text will be inferred from the containing class'
            arguments' section, if it exists. Alternatively, you can directly supply
            the help text here.
        parse: Optional function(s) which accepts the raw argument as input and
            return the parsed value. This can be either a single function, or a sequence of functions.
            If a sequence is provided, they will be called (in order) with their return value
            chained into the next function in the sequence.
        parse_inference: Whether to include the default inferred parser for the annotated.

        group: Optional group names for the argument. This affects how they're displayed
            in the backend's help text. Note this can also be a `Group` instance in order
            to control group order and/or group exclusivity.
        hidden: Whether the argument should be hidden in help text. Defaults to False.

        action: Generally automatically inferred from the data type. This allows to
            override the default.
        num_args: Generally automatically inferred from the data type. This allows to
            override the default.
        choices: Generally automatically inferred from the data type. This allows to
            override the default.
        completion: Used to provide custom completions. If specified, should be a function
            which accepts a partial string value and returns a list of
            [cappa.Completion](cappa.Completion) objects.
        required: Defaults to automatically inferring requiredness, based on whether the
            class's value has a default. By setting this, you can force a particular value.
        field_name: The name of the class field to populate with this arg. In most usecases,
            this field should be left unspecified and automatically inferred.
        deprecated: If supplied, the argument will be marked as deprecated. If given `True`,
            a default message will be generated, otherwise a supplied string will be
            used as the deprecation message.
        show_default: Whether to show the default value in help text. Defaults to `True`.
        destructure: When set, destructures the annotated type into current-level arguments.
            Generally prefer to use `field: Destructured[T]` syntax to setting this directly.
            See `Arg.destructure`.
        has_value: Whether the argument has a value that should be saved back to the destination
            type. For most `Arg`, this will default to `True`, however `--help` is an example
            of an `Arg` for which it is false.
        propagate: Specifies that an argument can be matched to all child. Global arguments only
            propagate down. When used at the top-level, in effect it creates a "global" argument.
    """

    def __hash__(self):
        return id(self)

    value_name: str | EmptyType = Empty
    short: bool | str | list[str] | None = False
    long: bool | str | list[str] | None = False
    count: bool = False
    default: Default | T | EmptyType | None = Empty
    help: str | None = None
    parse: Callable[..., T] | Sequence[Callable[..., Any]] | None = None
    parse_inference: bool = True

    group: str | tuple[int, str] | Group | EmptyType = Empty

    hidden: bool = False
    action: ArgActionType | None = None
    num_args: int | NumArgs | None = None
    choices: list[str] | None = None
    completion: Callable[..., list[Completion]] | None = None
    required: bool | None = None
    field_name: str | EmptyType = Empty
    deprecated: bool | str = False
    show_default: bool | str | DefaultFormatter = True
    propagate: bool = False

    destructure: Destructure | bool | None = None
    has_value: bool | None = None

    type_view: TypeView[Any] | None = None

    @classmethod
    def collect(
        cls,
        field: Field,
        type_view: TypeView[Any],
        fallback_help: str | None = None,
        default_short: bool = False,
        default_long: bool = False,
        state: State[Any] | None = None,
    ) -> list[FinalArg[Any] | FinalDestructure[Any]]:
        args: list[Arg[Any]] = find_annotations(type_view, cls) or [Arg()]

        exclusive = field.name if len(args) > 1 else False

        docs: list[DocType] = find_annotations(type_view, Doc)
        fallback_help = docs[0].documentation if docs else fallback_help

        destructure_: list[Destructure] = find_annotations(type_view, Destructure)
        destructure_annotation: Destructure | None = (
            destructure_[0] if destructure_ else None
        )

        # Dataclass field metadata takes precedence if it exists.
        field_metadata: list[Arg[Any]] = extract_dataclass_metadata(field, Arg)
        if field_metadata:
            args = field_metadata

        result: list[FinalArg[Any] | FinalDestructure[Any]] = []
        for i, arg in enumerate(args, start=1):
            # field_name and default are evaluated outside `normalize` because they can
            # potentially depend on the type's dataclass-like field information, whereas
            # nothing else does.
            field_name = infer_field_name(arg, field)
            default = infer_default(arg, type_view, field, destructure_annotation)

            normalized_arg = arg.normalize(
                type_view,
                fallback_help=fallback_help,
                default_short=default_short,
                default_long=default_long,
                exclusive=exclusive,
                field_name=field_name,
                default=default,
                state=state,
                destructure=destructure_annotation,
                show_default=arg.show_default if i == len(args) else False,
            )

            if normalized_arg.destructure:
                result.extend(normalized_arg.destructure.explode_args())
            else:
                result.append(normalized_arg)

        return list(explode_negated_bool_args(result))

    def normalize(
        self,
        type_view: TypeView[Any] | None = None,
        fallback_help: str | None = None,
        action: ArgActionType | None = None,
        default: Any = Empty,
        field_name: str | None = None,
        default_short: bool = False,
        default_long: bool = False,
        exclusive: bool | str = False,
        state: State[Any] | None = None,
        destructure: Destructure | bool | None = None,
        show_default: bool | str | DefaultFormatter | None = None,
    ) -> FinalArg[Any]:
        if type_view is None:
            type_view = TypeView(Any)

        default = Default.fallback_from(default, self.default)

        field_name = cast(str, field_name or self.field_name)

        short = infer_short(self, field_name, default_short)
        long = infer_long(self, type_view, field_name, default_long)
        choices = infer_choices(self, type_view)
        action = action or infer_action(self, type_view, long, default)
        num_args = NumArgs.infer(
            type_view, field_name, arg=self, action=action, long=long
        )

        required = infer_required(self.required, num_args, self.num_args, default)

        parse = infer_parse(self, type_view, state=state)
        help = infer_help(self, fallback_help)
        completion = infer_completion(self, choices)

        group = Group.infer(self, bool(short or long), exclusive)

        value_name = infer_value_name(self, field_name, num_args)
        has_value = infer_has_value(self, action)

        if self.propagate and not short and not long:
            raise ValueError(
                "`Arg.propagate` requires a non-positional named option (`short` or `long`)."
            )

        default_formatter: DefaultFormatter = DefaultFormatter.from_unknown(
            show_default if show_default is not None else self.show_default
        )

        destructure = Destructure.collect(
            field_name,
            default,
            destructure or self.destructure,
            type_view,
            default_short=default_short,
            default_long=default_long,
        )
        result: FinalArg[Any] = FinalArg(
            # preserved from self
            count=self.count,
            hidden=self.hidden,
            deprecated=self.deprecated,
            propagate=self.propagate,
            parse_inference=self.parse_inference,
            # computed/narrowed
            value_name=value_name,
            short=short,
            long=long,
            default=default,
            help=help,
            parse=parse,
            group=group,
            action=action,
            num_args=num_args,
            choices=choices,
            completion=completion,
            required=required,
            field_name=field_name,
            show_default=default_formatter,
            destructure=destructure,
            has_value=has_value,
            type_view=type_view,
        )
        verify_type_compatibility(
            result, field_name, type_view, has_custom_parse=bool(self.parse)
        )
        return result

    @classmethod
    def destructured(cls):
        """Mark a an argument as destructured. See also the shorter `Destructured` alias."""
        return cls(destructure=Destructure())


@dataclasses.dataclass(frozen=True)
class FinalArg(Arg[T]):
    """Post-normalization form of :class:`Arg` with narrowed field types.

    Produced exclusively by :meth:`Arg.normalize`. After normalization all
    previously-optional/Empty fields are guaranteed to be concrete values.
    """

    value_name: str = ""
    short: list[str] | Literal[False] = False
    long: list[str] | Literal[False] = False
    default: Default = dataclasses.field(default_factory=Default)
    group: Group = dataclasses.field(default_factory=Group)
    action: ArgActionType = ArgAction.set
    num_args: NumArgs = dataclasses.field(default_factory=NumArgs)
    required: bool = False
    field_name: str = ""
    has_value: bool = True
    type_view: TypeView[Any] = dataclasses.field(default_factory=lambda: TypeView(Any))
    parse: Callable[..., Any] = dataclasses.field(default=parse_value)
    show_default: DefaultFormatter = dataclasses.field(default_factory=DefaultFormatter)
    destructure: FinalDestructure[Any] | None = None

    def map_result(
        self,
        prog: str,
        parsed_args: dict[str, Any],
        state: State[Any] | None = None,
        input: TextIO | None = None,
    ) -> Resolved[T]:
        field_name = self.field_name
        if field_name in parsed_args:
            is_parsed, value = False, parsed_args[field_name]
        else:
            is_parsed, value = self.default(state=state, input=input)
        handler = parse_handler(self.parse, prog, value, self.names_str())
        return Resolved(handler, args=(value, is_parsed))

    def names(self, *, n: int = 0) -> list[str]:
        result = (self.short or []) + (self.long or [])
        return result[:n] if n else result

    def names_str(self, delimiter: str = ", ", *, n: int = 0) -> str:
        if self.short or self.long:
            return delimiter.join(self.names(n=n))
        return self.value_name

    @cached_property
    def is_option(self) -> bool:
        return bool(self.short or self.long)


def verify_type_compatibility(
    arg: FinalArg[Any],
    field_name: str,
    type_view: TypeView[Any],
    *,
    has_custom_parse: bool = False,
):
    """Verify classes of annotations are compatible with one another.

    Thus far:
        * Sequence and scalar types should not be mixed, unless an explicit `parse`
          is provided. Otherwise, they will always result in sequence-type results
          which will not map correctly to scalar types.

        * num_args!=1 should only be used with sequence types.
        * ArgAction.append should only be used with sequence types.
        * Non-required argument values with a None default must be Optional in the type annotation.
    """
    action = arg.action
    if has_custom_parse or ArgAction.is_custom(action):
        return

    num_args = arg.num_args
    if not num_args.required and num_args.default is None and not type_view.is_optional:
        raise ValueError(
            f"On field '{field_name}', `NumArgs(required=False, default=None)` requires the "
            f"annotated type to include `None` (e.g. `{type_view.repr_type} | None`), "
            "since the absent-value default of `None` is not assignable to the field type."
        )

    if type_view.is_union:
        all_same_arity = {
            ta.is_subclass_of((list, tuple, set))
            for ta in type_view.strip_optional().inner_types
        }
        if len(all_same_arity) > 1:
            raise ValueError(
                f"On field '{field_name}', apparent mismatch of annotated type with `Arg` options. "
                'Unioning "sequence" types with non-sequence types is not currently supported, '
                "unless using `Arg(parse=...)` or `Arg(action=<callable>)`. "
                "See [documentation](https://cappa.readthedocs.io/en/latest/annotation.html) for more details."
            )
        return

    if type_view.is_subclass_of((list, tuple, set)):
        if num_args.n in {0, 1} and action not in {ArgAction.append, None}:
            raise ValueError(
                f"On field '{field_name}', apparent mismatch of annotated type with `Arg` options. "
                f"'{type_view.repr_type}' type produces a sequence, whereas `num_args=1`/`action={action}` do not. "
                "See [documentation](https://cappa.readthedocs.io/en/latest/annotation.html) for more details."
            )
    else:
        if num_args.n not in {None, 0, 1} or action is ArgAction.append:
            raise ValueError(
                f"On field '{field_name}', apparent mismatch of annotated type with `Arg` options. "
                f"'{type_view.repr_type}' type produces a scalar, whereas `num_args={num_args}`/`action={action}` do not. "
                "See [documentation](https://cappa.readthedocs.io/en/latest/annotation.html) for more details."
            )


def infer_field_name(arg: Arg[Any], field: Field) -> str:
    if arg.field_name is not Empty:
        raise ValueError("Arg 'name' cannot be set when using automatic inference.")

    return field.name


def infer_default(
    arg: Arg[Any],
    type_view: TypeView[Any],
    field: Field | None = None,
    destructure: Destructure | None = None,
) -> Default:
    if field:
        inferred_default: Default = infer_default(
            arg, type_view, destructure=destructure
        )

        if field.default is not Empty:
            return inferred_default | field.default

        if field.default_factory is not Empty:
            return inferred_default | ValueFrom(field.default_factory)

        return inferred_default

    default = arg.default
    if default is not Empty:
        return Default.from_value(default)

    if type_view.is_optional:
        return Default(default=None)

    if destructure:
        return Default(ValueFrom(type_view.annotation))

    if type_view.is_subclass_of(bool):
        return Default(default=False)

    return Default()


def infer_required(
    required: bool | None,
    num_args: NumArgs,
    raw_num_args: int | NumArgs | None,
    default: Default,
):
    if required is True:
        return True

    if default.has_value:
        if required is False:
            raise ValueError(
                "When specifying `required=False`, a default value must be supplied able to be "
                "supplied through type inference, `Arg(default=...)`, or through class-level default"
            )

        # Only yield to "unbounded = optional" when the user *explicitly* set a negative
        # num_args (e.g. Arg(num_args=-1) or Arg(num_args=NumArgs(n=-1))).  An inferred
        # -1 (e.g. from list[str]) keeps the arg required (at-least-one semantics).
        raw_n = NumArgs.raw_n(raw_num_args)
        if raw_n is not None and raw_n < 0:
            return False

        return True
    return False


def infer_short(
    arg: Arg[Any], name: str, default: bool = False
) -> list[str] | Literal[False]:
    short = arg.short or default
    if not short:
        return False

    if isinstance(short, bool):
        short_name = name[0]
        return [f"-{short_name}"]

    if isinstance(short, str):
        short = short.split("/")
    else:
        short = short

    return [item if item.startswith("-") else f"-{item}" for item in short]


def infer_long(
    arg: Arg[Any], type_view: TypeView[Any], name: str, default: bool
) -> list[str] | Literal[False]:
    long = arg.long or default

    if not long:
        # bools get automatically coerced into flags, otherwise stay off.
        if not type_view.is_subclass_of(bool):
            return False

        long = True

    if isinstance(long, bool):
        long = name.replace("_", "-")
        return [f"--{long}"]

    if isinstance(long, str):
        long = long.split("/")

    return [item if item.startswith("--") else f"--{item}" for item in long]


def infer_choices(arg: Arg[Any], type_view: TypeView[Any]) -> list[str] | None:
    if arg.choices is None:
        choices = detect_choices(type_view)
    else:
        choices = arg.choices

    if not choices:
        return None

    return [str(c) for c in choices]


def infer_action(
    arg: Arg[Any],
    type_view: TypeView[Any],
    long: list[str] | Literal[False],
    default: Default,
) -> ArgActionType:
    if arg.count:
        return ArgAction.count

    if arg.action is not None:
        return arg.action

    type_view = type_view.strip_optional()

    # Coerce raw `bool` into flags by default
    if type_view.is_subclass_of(bool):
        if default.default is not Empty and bool(default.default):
            return ArgAction.store_false

        return ArgAction.store_true

    has_custom_parse = arg.parse is not None
    has_custom_num_args = arg.num_args is not None

    is_positional = not arg.short and not long
    raw_n = NumArgs.raw_n(arg.num_args)
    is_unbounded = raw_n == -1

    if (
        has_custom_parse
        or is_unbounded
        or (is_positional and not has_custom_num_args)
        or (has_custom_num_args and raw_n != 1)
    ):
        return ArgAction.set

    if type_view.is_subtype_of((list, set)):
        return ArgAction.append

    if type_view.is_variadic_tuple:
        return ArgAction.append

    return ArgAction.set


def infer_parse(
    arg: Arg[Any], type_view: TypeView[Any], state: State[Any] | None = None
) -> Callable[..., Any]:
    parse: Parser[Any] | Sequence[Parser[Any]] = arg.parse  # type: ignore
    default_parse = parse_value(type_view)

    # This is the original arg choices, e.g. explicitly provided. Inferred choices
    # need to be handled internally to the parsers.
    parsers: Sequence[Parser[Any]] = []
    if parse:
        if isinstance(parse, Sequence):
            parsers = [*cast(Sequence[Any], parse)]
        else:
            parsers = [parse]

    if arg.choices:
        literal_type = Literal[tuple(arg.choices)]  # type: ignore
        literal_parse: Parser[Any] = parse_literal(literal_type)  # type: ignore
        parsers = cast(Sequence[Parser[Any]], [*parsers, literal_parse])

    if arg.parse_inference:
        parsers = [*parsers, default_parse]

    return evaluate_parse(parsers, type_view, state=state)


def infer_help(arg: Arg[Any], fallback_help: str | None) -> str | None:
    help = arg.help
    if help is None:
        help = fallback_help

    return help


def infer_completion(
    arg: Arg[Any], choices: list[str] | None
) -> Callable[[str], list[Completion]] | None:
    if arg.completion:
        return arg.completion

    if choices:
        return complete_choices(choices, help=arg.help)

    return None


def infer_value_name(arg: Arg[Any], field_name: str, num_args: NumArgs) -> str:
    if arg.value_name is not Empty:
        return arg.value_name

    if num_args.n > 1:
        return " ".join([field_name] * num_args.n)

    return field_name


def explode_negated_bool_args(
    args: Sequence[FinalArg[Any] | FinalDestructure[Any]],
) -> Iterable[FinalArg[Any] | FinalDestructure[Any]]:
    """Expand `--foo/--no-foo` solo arguments into dual-arguments.

    Acts as a transform from `Arg(long='--foo/--no-foo')` to
    `Annotated[Arg(long='--foo', action=ArgAction.store_true), Arg(long='--no-foo', action=ArgAction.store_false)]`
    """
    for arg in args:
        # Only args can *be* bool negatable args.
        if not isinstance(arg, FinalArg):
            yield arg
            continue

        yielded = False
        if isinstance(arg.action, ArgAction) and arg.action.is_bool_action and arg.long:
            long = arg.long

            negatives = [item for item in long if "--no-" in item]
            positives = [item for item in long if "--no-" not in item]
            if negatives and positives:
                assert isinstance(arg.default, Default)

                disabled: DefaultFormatter = DefaultFormatter.disabled()
                group = dataclasses.replace(
                    arg.group,
                    exclusive=True,
                    id=arg.field_name,
                )

                positive_arg = dataclasses.replace(
                    arg,
                    long=positives,
                    action=ArgAction.store_true,
                    show_default=disabled,
                    group=group,
                    help=None,
                )
                negative_arg = dataclasses.replace(
                    arg,
                    long=negatives,
                    action=ArgAction.store_false,
                    show_default=arg.show_default,
                    group=group,
                )

                yield positive_arg
                yield negative_arg
                yielded = True

        if not yielded:
            yield arg


def infer_has_value(arg: Arg[Any], action: ArgActionType):
    if arg.has_value is not None:
        return arg.has_value

    if isinstance(action, ArgAction) and action in ArgAction.meta_actions():
        return False

    return True


from cappa.destructure import Destructure, FinalDestructure  # noqa: E402



from __future__ import annotations

import dataclasses
from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import parse


@dataclass
class Items:
    items: Annotated[list[str], cappa.Arg(long=True, action=cappa.ArgAction.append)] = (
        dataclasses.field(default_factory=lambda: [])
    )


@dataclass
class Args:
    w: cappa.Destructured[Items]


def test_append_multiple():
    result = parse(Args, "--items=a", "--items=b", "--items=c")
    assert result == Args(w=Items(items=["a", "b", "c"]))


def test_append_default_empty():
    result = parse(Args)
    assert result == Args(w=Items(items=[]))



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import parse


@dataclass
class RequiredMissingOne:
    foo: Annotated[Union[str, None], cappa.Arg(long=True)] = None


@dataclass
class RequiredMissingTwo:
    foo: Annotated[Union[str, None], cappa.Arg(long=True)] = None


@dataclass
class RequiredMissing:
    subcommand: Annotated[
        Union[RequiredMissingOne, RequiredMissingTwo], cappa.Subcommand
    ]


def test_has_possible_values():
    with pytest.raises(cappa.Exit) as e:
        parse(RequiredMissing, "req", backend=None)
    assert isinstance(e.value.message, str)

    expected_message = (
        "Invalid command 'req' (Did you mean: "
        "[cappa.subcommand]required-missing-one[/cappa.subcommand], "
        "[cappa.subcommand]required-missing-two[/cappa.subcommand])"
    )
    assert expected_message == e.value.message


def test_no_possible_values():
    with pytest.raises(cappa.Exit) as e:
        parse(RequiredMissing, "bad", backend=None)
    assert isinstance(e.value.message, str)

    expected_message = "Invalid command 'bad'"
    assert expected_message == e.value.message



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Union

from typing_extensions import Annotated

import cappa
from cappa.parse import unpack_arguments
from tests.utils import Backend, backends, parse


@dataclass
class Single:
    value: str


@dataclass
class Double:
    one: str
    two: str


@dataclass
class Mapping:
    foo: str
    bar: str


@dataclass
class Args:
    single: Annotated[
        Union[Single, None], cappa.Arg(long=True, parse=unpack_arguments)
    ] = None
    double: Annotated[
        Union[Double, None],
        cappa.Arg(long=True, num_args=2, parse=unpack_arguments),
    ] = None
    mapping: Annotated[
        Union[Mapping, None],
        cappa.Arg(long=True, parse=[json.loads, unpack_arguments]),
    ] = None


@backends
def test_dataclass_annotation(backend: Backend):
    """Dataclass annotation will supply or splat arguments, depending on input shape."""
    result = parse(Args, backend=backend)
    assert result == Args()

    result = parse(Args, "--single=3", backend=backend)
    assert result.single == Single("3")
    assert result.double is None
    assert result.mapping is None

    result = parse(Args, "--double", "5", "10", backend=backend)
    assert result.single is None
    assert result.double == Double(one="5", two="10")
    assert result.mapping is None

    result = parse(Args, '--mapping={"foo": "8", "bar": "3"}', backend=backend)
    assert result.single is None
    assert result.double is None
    assert result.mapping == Mapping(foo="8", bar="3")



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class Config:
    path: str

    @classmethod
    def from_classmethod(cls, path: str):
        return cls(path)

    def from_method(self, path: str) -> str:
        return "/".join([self.path, path])


config = Config("foo")


@backends
def test_from_classmethod(backend: Backend):
    @cappa.command(name="command")
    @dataclass
    class Command:
        config: Annotated[Config, cappa.Arg(parse=Config.from_classmethod)]

    test = parse(Command, "foo", backend=backend)
    assert test == Command(config=Config("foo"))


@backends
def test_method(backend: Backend):
    @cappa.command(name="command")
    @dataclass
    class Command:
        config: Annotated[str, cappa.Arg(parse=config.from_method)]

    test = parse(Command, "bar", backend=backend)
    assert test == Command(config="foo/bar")



from __future__ import annotations

from dataclasses import dataclass

from tests.utils import Backend, backends, parse


@backends
def test_positional_with_default_is_optional(backend: Backend):
    @dataclass
    class ArgTest:
        arg: int = 0

    result = parse(ArgTest, backend=backend)
    assert result.arg == 0


@backends
def test_multiple_positionals_fill_in_order(backend: Backend):
    @dataclass
    class ArgTest:
        arg: int = 0
        arg2: int = 0
        arg3: int = 0

    result = parse(ArgTest, "3", "5", backend=backend)
    assert result.arg == 3
    assert result.arg2 == 5
    assert result.arg3 == 0



from __future__ import annotations

import os
from pathlib import Path
from typing import Any, Iterable, cast

from cappa.arg import Arg
from cappa.command import FinalCommand
from cappa.completion.shells import available_shells
from cappa.output import Exit, Output
from cappa.parser import Completion, FileCompletion, backend


def execute(
    command: FinalCommand[Any],
    prog: str,
    action: str,
    arg: Arg[Any],
    output: Output,
):
    shell_name = Path(os.environ.get("SHELL", "bash")).name
    shell = available_shells.get(shell_name)

    if shell is None:
        raise Exit("Unknown shell", code=1)

    if action == "generate":
        raise Exit(shell.backend_template(prog, arg), code=0)

    command_args = parse_incomplete_command()

    backend(
        command,
        command_args,
        output=output,
        prog=prog,
        provide_completions=True,
    )


def parse_incomplete_command() -> list[str]:
    raw_completion_line = os.environ.get("COMPLETION_LINE", "").replace("\n", " ")
    raw_completion_location = os.environ.get("COMPLETION_LOCATION")
    if not raw_completion_line or raw_completion_location is None:
        raise Exit(code=0)

    completion_line = list(split_incomplete_command(raw_completion_line))
    completion_index = int(raw_completion_location)

    return completion_line[1:completion_index]


def split_incomplete_command(string: str) -> Iterable[str]:
    import shlex

    lex = shlex.shlex(string, posix=True)
    lex.whitespace_split = True
    lex.commenters = ""

    try:
        yield from lex
    except ValueError:
        yield lex.token


def format_completions(*completions: Completion | FileCompletion) -> str | None:
    if not completions:
        return None

    if isinstance(completions[0], FileCompletion):
        return "file"

    result: list[str] = []
    for item in completions:
        item = cast(Completion, item)
        result.append(f"{item.value}:{item.description}")
    return "\n".join(result)



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class Command:
    foo: Annotated[int, cappa.Arg(default=4)]
    bar: Annotated[int, cappa.Arg(default=5)]


@backends
def test_valid(backend: Backend):
    test = parse(Command, "1", "2", backend=backend)
    assert test == Command(1, 2)


@backends
def test_default_is_not_mapped(backend: Backend):
    @dataclass
    class Command:
        foo: int = "4"  # type: ignore

    test = parse(Command, "1", backend=backend)
    assert test == Command(1)

    test = parse(Command, backend=backend)
    assert test == Command("4")  # type: ignore


@backends
def test_explicit_arg_default_precedes_class_level(backend: Backend):
    @dataclass
    class Command:
        foo: Annotated[int, cappa.Arg(default=10)] = 4

    test = parse(Command, "1", backend=backend)
    assert test == Command(1)

    test = parse(Command, backend=backend)
    assert test == Command(10)



from __future__ import annotations

import cappa
from tests.utils import Backend, backends, parse


@cappa.command(name="sub")
class Example:
    bar: int
    name: str


@backends
def test_invoke_top_level_command(backend: Backend):
    result = parse(Example, "4", "foo", backend=backend)
    assert result == Example(bar=4, name="foo")  # pyright: ignore


@backends
def test_no_args_command(backend: Backend):
    @cappa.command
    class Example:
        bar: int
        name: str

    result = parse(Example, "4", "foo", backend=backend)
    assert result == Example(bar=4, name="foo")



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_invalid_choice_help(backend: Backend):
    @dataclass
    class Args:
        arg: str
        option: Annotated[str, cappa.Arg(long=True)]
        arg2: str
        option2: Annotated[str, cappa.Arg(long=True)]

    result = parse(
        Args, "arg", "--option=opt", "arg2", "--option2=opt2", backend=backend
    )
    assert result == Args(arg="arg", option="opt", arg2="arg2", option2="opt2")



from __future__ import annotations

from dataclasses import dataclass
from unittest.mock import patch

import pytest
from typing_extensions import Annotated

import cappa
from cappa.default import ValueFrom
from tests.utils import Backend, backends, parse


def woah(value: int) -> int:
    return 1 + value


@dataclass
class Command:
    foo: Annotated[int, cappa.Arg(default=cappa.Env("FOO"))] = 4
    bar: Annotated[
        int, cappa.Arg(default=cappa.Env("BAR") | ValueFrom(woah, value=4))
    ] = 0


@backends
def test_valid(backend: Backend):
    result = parse(Command, "6", "7", backend=backend)
    assert result == Command(6, 7)

    result = parse(Command, backend=backend)
    assert result == Command(4, 5)

    with patch("os.environ", new={"FOO": "3", "BAR": "9"}):
        result = parse(Command, backend=backend)
        assert result == Command(3, 9)


def raises():
    raise ValueError()


@backends
def test_not_parsed(backend: Backend):
    @dataclass
    class Command:
        foo: Annotated[int, cappa.Arg(parse=raises, default=ValueFrom(int))]

    with pytest.raises(cappa.Exit):
        parse(Command, "7", backend=backend)

    result = parse(Command, backend=backend)
    assert result == Command(0)


def called():
    return ["42"]


@backends
def test_positional_unbound_num_args(backend: Backend):
    @dataclass
    class Command:
        positional: Annotated[
            list[str],
            cappa.Arg(
                short=False,
                long=False,
                default=cappa.ValueFrom(called),
            ),
        ]

    result = parse(Command, "7", backend=backend)
    assert result == Command(["7"])

    result = parse(Command, backend=backend)
    assert result == Command(["42"])


def two_tuple():
    return ("42", "0")


def test_optional_positional_2_tuple():
    @dataclass
    class Command:
        positional: Annotated[
            tuple[str, str],
            cappa.Arg(
                short=False,
                long=False,
                default=cappa.ValueFrom(two_tuple),
            ),
        ]

    result = parse(Command, "7", "8")
    assert result == Command(("7", "8"))

    result = parse(Command)
    assert result == Command(("42", "0"))



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, TypedDict, Union

from typing_extensions import Annotated

import cappa
from cappa.output import Output
from cappa.state import State
from tests.utils import Backend, CapsysOutput, backends, invoke, parse


@backends
def test_no_op(backend: Backend):
    """Base case, this doesnt really serve any purpose."""

    def function():
        return 4

    parse(function, backend=backend)


@backends
def test_parse_one_arg(backend: Backend):
    """Parse with arguments.

    Again, not really a logical usecase for `parse`, but technically it ought to work.
    """

    def function(foo: str):
        return foo

    result = parse(function, "meow", backend=backend)
    assert result.foo == "meow"  # pyright: ignore


@backends
def test_invoke_base_case(backend: Backend):
    def function(foo: str):
        return foo

    result = invoke(function, "meow", backend=backend)
    assert result == "meow"


@backends
def test_invoke_optional_args(backend: Backend):
    def function(foo: Annotated[int, cappa.Arg(long=True)] = 15):
        return foo

    result = invoke(function, backend=backend)
    assert result == 15

    result = invoke(function, "--foo", "53", backend=backend)
    assert result == 53


@dataclass
class Sub:
    bar: Annotated[int, cappa.Arg(long=True)]

    def __call__(self):
        return self.bar + 1


@backends
def test_subcommand(backend: Backend):
    def function(sub: cappa.Subcommands[Union[Sub, None]] = None):
        assert sub is None
        return 6

    result = invoke(function, backend=backend)
    assert result == 6

    result = invoke(function, "sub", "--bar", "34", backend=backend)
    assert result == 35


def foo():
    return 5


@backends
def test_invoke_partial_arg_partial_dep(backend: Backend):
    def function(
        dep: Annotated[int, cappa.Dep(foo)],
        foo: Annotated[int, cappa.Arg(long=True)] = 15,
    ):
        return dep + foo

    result = invoke(function, backend=backend)
    assert result == 20

    result = invoke(function, "--foo", "53", backend=backend)
    assert result == 58


@backends
def test_output(backend: Backend, capsys: Any):
    """Depends on Output DI."""

    def function(foo: str, output: Output):
        output("hey")
        return foo

    result = invoke(function, "meow", backend=backend)
    assert result == "meow"

    out = CapsysOutput.from_capsys(capsys)
    assert out.stdout == "hey\n"


class StateFoo(TypedDict):
    foo: int


@backends
def test_state(backend: Backend, capsys: Any):
    """Depends on Output DI."""

    def function(foo: int, state: State[StateFoo]):
        return foo + state.state["foo"]

    state = State(StateFoo(foo=45))
    result = invoke(function, "3", backend=backend, state=state)
    assert result == 48



import sys

sys.path.insert(0, ".")

project = "docutils"
copyright = "2025, docutils"
author = "docutils"
release = "docutils"

extensions = [
    "cappa.ext.docutils",
]

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", ".venv"]
html_theme = "alabaster"



from __future__ import annotations

from dataclasses import dataclass

import pytest
from typing_extensions import Annotated, Literal

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class ArgTest:
    list: Annotated[
        list[Literal["one", "two", "three", 4]],
        cappa.Arg(short=True, default=[]),
    ]
    tuple: Annotated[
        tuple[Literal["one", "two", "three", 4], ...],
        cappa.Arg(short=True, default=()),
    ]
    set: Annotated[
        set[Literal["one", "two", "three", 4]],
        cappa.Arg(short=True, default=set()),  # pyright: ignore
    ]


@backends
def test_list(backend: Backend):
    test = parse(ArgTest, "-l", "one", "-l", "two", backend=backend)
    assert test.list == ["one", "two"]


@backends
def test_tuple(backend: Backend):
    test = parse(ArgTest, "-t", "one", "-t", "two", backend=backend)
    assert test.tuple == ("one", "two")


@backends
def test_set(backend: Backend):
    test = parse(ArgTest, "-s", "one", "-s", "two", backend=backend)
    assert test.set == {"one", "two"}


@backends
def test_invalid(backend: Backend):
    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "-l", "one", "-l", "wat", backend=backend)

    message = str(e.value.message).lower()
    assert "invalid choice: 'wat' (choose from 'one', 'two', 'three', 4)" in message



import pytest

from cappa.class_inspect import fields


def test_invalid_class_base():
    class Random: ...

    with pytest.raises(ValueError) as e:
        fields(Random)
    assert (
        "'test_invalid_class_base.<locals>.Random' is not a currently supported kind of class."
        in str(e.value)
    )



from __future__ import annotations

from dataclasses import dataclass

import cappa
from tests.utils import parse


@dataclass
class Foo:
    foo: int = 4


@dataclass
class CLI:
    attrs: cappa.Destructured[Foo | None] = None


def test_destructured_optional():
    result = parse(CLI)
    assert result == CLI(attrs=None)



from __future__ import annotations

from dataclasses import dataclass

from tests.utils import parse_completion


def test_long_option_name():
    @dataclass
    class Args:
        default: bool = False

    result = parse_completion(Args, "--default")
    assert result is None



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, invoke


def dependency():
    return 5


@dataclass
class Command:
    foo: int

    def __call__(self, dep: Annotated[int, cappa.Dep(dependency)]):
        return self.foo + dep


@backends
def test_invoke_top_level_command(backend: Backend):
    result = invoke(Command, "7", backend=backend)
    assert result == 7 + 5



from __future__ import annotations

import textwrap
from dataclasses import dataclass
from typing import Any, Literal

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import CapsysOutput, parse, strip_trailing_whitespace


@dataclass
class Args:
    required: Annotated[str, cappa.Arg(help="I'm required")]
    name: Annotated[str, cappa.Arg(help="I'm optional")] = "arg"
    short: Annotated[str, cappa.Arg(short=True, help="I'm an option")] = "opt"
    maybe: Annotated[str | None, cappa.Arg(short=True, help="maybe?")] = None


def test_default_settings(capsys: Any):
    with pytest.raises(cappa.HelpExit) as e:
        parse(Args, "--help", completion=False)

    assert e.value.code == 0
    out = strip_trailing_whitespace(capsys.readouterr().out)

    assert out == textwrap.dedent(
        """\
        Usage: args [-s SHORT] [-m MAYBE] REQUIRED [NAME] [-h]

          Options
            [-s SHORT]    I'm an option (Default: opt)
            [-m MAYBE]    maybe?

          Arguments
            REQUIRED      I'm required
            [NAME]        I'm optional (Default: arg)

          Help
            [-h, --help]  Show this message and exit.
        """
    )


def test_override_default_format(capsys: Any):
    with pytest.raises(cappa.HelpExit) as e:
        parse(
            Args,
            "--help",
            completion=False,
            help_formatter=cappa.HelpFormatter.default.with_default_format(
                "[Default '{default}']"
            ),
        )

    assert e.value.code == 0
    out = strip_trailing_whitespace(capsys.readouterr().out)

    assert out == textwrap.dedent(
        """\
        Usage: args [-s SHORT] [-m MAYBE] REQUIRED [NAME] [-h]

          Options
            [-s SHORT]    I'm an option [Default 'opt']
            [-m MAYBE]    maybe?

          Arguments
            REQUIRED      I'm required
            [NAME]        I'm optional [Default 'arg']

          Help
            [-h, --help]  Show this message and exit.
        """
    )


def test_override_help_format(capsys: Any):
    with pytest.raises(cappa.HelpExit) as e:
        parse(
            Args,
            "--help",
            completion=False,
            help_formatter=cappa.HelpFormatter.default.with_arg_format(
                (
                    "{default}",
                    "{help}",
                )
            ),
        )

    assert e.value.code == 0
    out = strip_trailing_whitespace(capsys.readouterr().out)

    assert out == textwrap.dedent(
        """\
        Usage: args [-s SHORT] [-m MAYBE] REQUIRED [NAME] [-h]

          Options
            [-s SHORT]    (Default: opt) I'm an option
            [-m MAYBE]    maybe?

          Arguments
            REQUIRED      I'm required
            [NAME]        (Default: arg) I'm optional

          Help
            [-h, --help]  Show this message and exit.
        """
    )


def test_choice_options(capsys: Any):
    @dataclass
    class Args:
        required: Annotated[Literal["one", "two", "three"], cappa.Arg(help="Required.")]

    with pytest.raises(cappa.HelpExit) as e:
        parse(
            Args,
            "--help",
            completion=False,
        )

    assert e.value.code == 0
    out = strip_trailing_whitespace(capsys.readouterr().out)

    assert out == textwrap.dedent(
        """\
        Usage: args REQUIRED [-h]

          Arguments
            REQUIRED      Required. Valid options: one, two, three.

          Help
            [-h, --help]  Show this message and exit.
        """
    )

    with pytest.raises(cappa.HelpExit) as e:
        parse(
            Args,
            "--help",
            completion=False,
            help_formatter=cappa.HelpFormatter().with_arg_format("{help}"),
        )

    assert e.value.code == 0
    out = strip_trailing_whitespace(capsys.readouterr().out)

    assert out == textwrap.dedent(
        """\
        Usage: args REQUIRED [-h]

          Arguments
            REQUIRED      Required.

          Help
            [-h, --help]  Show this message and exit.
        """
    )


def test_callable_help_formatter(capsys: Any):
    @dataclass
    class Args:
        required: Annotated[int, cappa.Arg(help="Required.")]
        not_required: Annotated[int | None, cappa.Arg(help="erm.")] = None

    def help_formatter(arg: cappa.Arg[Any]) -> str | None:
        if arg.field_name == "required":
            return f"Num args: {arg.num_args}"
        return None

    with pytest.raises(cappa.HelpExit) as e:
        parse(
            Args,
            "--help",
            completion=False,
            help_formatter=cappa.HelpFormatter().with_arg_format(
                (
                    "{help}",
                    help_formatter,
                )
            ),
        )

    assert e.value.code == 0
    out = strip_trailing_whitespace(capsys.readouterr().out)

    assert out == textwrap.dedent(
        """\
        Usage: args REQUIRED [NOT_REQUIRED] [-h]

          Arguments
            REQUIRED        Required. Num args: 1
            [NOT_REQUIRED]  erm.

          Help
            [-h, --help]    Show this message and exit.
        """
    )


def test_explicitly_wrapped_formatter(capsys: Any):
    @dataclass
    class Args:
        name: Annotated[str, cappa.Arg(help="Optional.")] = "arg"

    help_formatter = cappa.HelpFormatter(default_format="Default - {default}!")
    with pytest.raises(cappa.HelpExit) as e:
        parse(Args, "--help", help_formatter=help_formatter)

    assert e.value.code == 0
    output = CapsysOutput.from_capsys(capsys)
    assert "Optional. Default - arg!\n" in output.stdout



from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from unittest.mock import patch

import pytest

import cappa
from cappa.parser import backend
from tests.utils import parse


def test_unknown_shell():
    @dataclass
    class Args:
        value: str

    with patch("os.environ", new={"SHELL": "notbash"}):
        with pytest.raises(cappa.Exit) as e:
            parse(Args, "--completion", "complete", backend=backend)

    assert e.value.code == 1
    assert e.value.message == "Unknown shell"


def test_generate_completions():
    @dataclass
    class Args:
        value: str

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "--completion", "generate", backend=backend)

    assert e.value.code == 0
    assert e.value.message
    assert "_args_completion" in str(e.value.message)


def test_invalid_setup():
    @dataclass
    class Args:
        value: str

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "--completion", "complete", backend=backend)

    assert e.value.code == 0
    assert not e.value.message


def test_no_completion():
    @dataclass
    class Example: ...

    result = cappa.parse(Example, argv=[], backend=backend, completion=False)
    assert result == Example()

    with pytest.raises(cappa.Exit) as e:
        cappa.parse(
            Example,
            argv=["--completion", "generate"],
            backend=backend,
            completion=False,
        )
    assert e.value.code == 2


def test_arg_completion(capsys: Any):
    @dataclass
    class Example: ...

    completion: cappa.Arg[Any] = cappa.Arg(short="-p", long="--pompletion")

    result = cappa.parse(Example, argv=[], completion=completion, backend=backend)
    assert result == Example()

    with pytest.raises(SystemExit) as e:
        cappa.parse(
            Example,
            argv=["-p", "generate"],
            completion=completion,
            backend=backend,
        )
    assert e.value.code == 0

    out = capsys.readouterr().out
    assert "_example_completion" in out

    with pytest.raises(SystemExit) as e:
        cappa.parse(
            Example,
            argv=["--help"],
            completion=completion,
            backend=backend,
        )
    assert e.value.code == 0

    out = capsys.readouterr().out
    assert "-p, --pompletion" in out



from __future__ import annotations

import asyncio
from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from cappa.default import ValueFrom
from tests.utils import Backend, backends


async def async_list_provider() -> list[str]:
    await asyncio.sleep(0.001)
    return ["async", "list", "values"]


@dataclass
class Command:
    values: Annotated[
        list[str],
        cappa.Arg(
            short=False,
            long=False,
            default=ValueFrom(async_list_provider),
        ),
    ]

    def __call__(self):
        return self


@backends
def test(backend: Backend):
    result = asyncio.run(cappa.invoke_async(Command, argv=[], backend=backend))
    assert result.values == ["async", "list", "values"]

    result = asyncio.run(cappa.invoke_async(Command, argv=["a", "b"], backend=backend))
    assert result.values == ["a", "b"]



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Any

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, invoke

log = logging.getLogger("test")


def level_three():
    log.debug("three")
    return {"three": True}


def level_two(three: Annotated[dict[str, Any], cappa.Dep(level_three)]):
    log.debug("two")
    return {"two": {**three}}


def level_one(
    two: Annotated[dict[str, Any], cappa.Dep(level_two)],
    three: Annotated[dict[str, Any], cappa.Dep(level_three)],
):
    log.debug("one")
    return {"one": {**two, **three}}


def command(
    one: Annotated[dict[str, Any], cappa.Dep(level_one)],
    two: Annotated[dict[str, Any], cappa.Dep(level_two)],
    three: Annotated[dict[str, Any], cappa.Dep(level_three)],
):
    return {**one, **two, **three}


@cappa.command(invoke=command)
@dataclass
class Command: ...


@backends
def test_invoke_top_level_command(caplog: Any, backend: Backend):
    """Assert multiple paths to explicit dependencies are still fulfilled.

    * Ensure they are only called once overall, regardless of the number of downstream dependents.
    """
    caplog.set_level(logging.DEBUG)

    result = invoke(Command, backend=backend)

    assert result == {
        "one": {"two": {"three": True}, "three": True},
        "two": {"three": True},
        "three": True,
    }

    assert len(caplog.records) == 3



from __future__ import annotations

from dataclasses import dataclass
from unittest.mock import patch

from typing_extensions import Annotated

import cappa
from cappa import Env, ValueFrom
from tests.utils import parse


@dataclass
class Foo:
    color: Annotated[str, cappa.Arg(long=True, default=Env("FOO"))]


@dataclass
class Args:
    attrs: cappa.Destructured[Foo]


def test_env_default_no_cli_arg():
    with patch("os.environ", new={"FOO": "blue"}):
        result = parse(Args)
    assert result == Args(attrs=Foo("blue"))


def test_env_default_cli_overrides():
    with patch("os.environ", new={"FOO": "blue"}):
        result = parse(Args, "--color=red")
    assert result == Args(attrs=Foo("red"))


@dataclass
class Bar:
    x: Annotated[int, cappa.Arg(long=True, default=Env("BAR_X"))]
    y: Annotated[str, cappa.Arg(long=True, default="fallback")]


@dataclass
class PartialArgs:
    bar: cappa.Destructured[Bar]


def test_partial_args_missing_field_uses_env():
    with patch("os.environ", new={"BAR_X": "42"}):
        result = parse(PartialArgs, "--y=hello")
    assert result == PartialArgs(bar=Bar(x=42, y="hello"))


def test_partial_args_all_from_env_and_default():
    with patch("os.environ", new={"BAR_X": "7"}):
        result = parse(PartialArgs)
    assert result == PartialArgs(bar=Bar(x=7, y="fallback"))


@dataclass
class Baz:
    val: Annotated[int, cappa.Arg(long=True, default=ValueFrom(lambda: 99))]


@dataclass
class ValueFromArgs:
    baz: cappa.Destructured[Baz]


def test_value_from_default():
    result = parse(ValueFromArgs)
    assert result == ValueFromArgs(baz=Baz(val=99))



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_negative_value_name_as_choice(backend: Backend):
    @dataclass
    class Args:
        level: Literal["-4", "-2", "0", "2", "4"]

    result = parse(Args, "-4", backend=backend)
    assert result == Args(level="-4")


@backends
def test_negative_value_name_with_option(backend: Backend):
    @dataclass
    class Args:
        verbose: Annotated[bool, cappa.Arg(short=True, default=False)]
        level: Literal["-4", "-2", "0", "2", "4"]

    result = parse(Args, "-v", "-4", backend=backend)
    assert result == Args(verbose=True, level="-4")


@backends
def test_negative_value_name_string_arg(backend: Backend):
    @dataclass
    class Args:
        option: Annotated[str, cappa.Arg(long=True)]
        level: Literal["-4", "-2", "0"]

    result = parse(Args, "--option", "foo", "-4", backend=backend)
    assert result == Args(option="foo", level="-4")



from __future__ import annotations

import asyncio
from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends


async def async_parse_int(value: str) -> int:
    await asyncio.sleep(0.001)
    result = int(value)
    if result < 0:
        raise ValueError("Value must be non-negative")
    return result


@dataclass
class Command:
    value: Annotated[int, cappa.Arg(parse=async_parse_int, default=100)]

    def __call__(self):
        return self


@backends
def test(backend: Backend):
    result = asyncio.run(cappa.invoke_async(Command, argv=[], backend=backend))
    assert result.value == 100

    result = asyncio.run(cappa.invoke_async(Command, argv=["42"], backend=backend))
    assert result.value == 42



from __future__ import annotations

from dataclasses import dataclass
from textwrap import dedent
from typing import Any

import pytest

import cappa
from tests.utils import (
    Backend,
    backends,
    parse,
    strip_trailing_whitespace,
    terminal_width,
)


@dataclass
class Flags:
    flag: bool = False
    enabled: bool = True


@dataclass
class Args:
    flags: cappa.Destructured[Flags]


@backends
def test_bool_destructured(backend: Backend, capsys: Any):
    assert parse(Args, "--flag") == Args(flags=Flags(flag=True, enabled=True))
    assert parse(Args, "--enabled") == Args(flags=Flags(flag=False, enabled=False))

    with terminal_width(80), pytest.raises(cappa.Exit):
        parse(Args, "--help", backend=backend, completion=False)

    result = strip_trailing_whitespace(capsys.readouterr().out)
    assert result == dedent(
        """\
        Usage: args [--flag] [--enabled] [-h]

          Options
            [--flag]      (Default: False)
            [--enabled]   (Default: True)

          Help
            [-h, --help]  Show this message and exit.
        """
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

from typing_extensions import Annotated

import cappa
from tests.utils import parse


@dataclass
class Args:
    flavors: list[Literal["chocolate", "vanilla"]]
    something: Annotated[int, cappa.Arg(short="-p")] = 9


def test_destructured():
    test = parse(
        Args,
        "chocolate",
        "vanilla",
        "-p4",
    )
    assert test == Args(
        flavors=["chocolate", "vanilla"],
        something=4,
    )



from __future__ import annotations

import asyncio
from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends


async def async_parse_multiply(value: str) -> int:
    await asyncio.sleep(0.001)
    return int(value) * 2


@dataclass
class Command:
    value: Annotated[int, cappa.Arg(parse=async_parse_multiply)]


@backends
def test(backend: Backend):
    result = asyncio.run(cappa.parse_async(Command, argv=["5"], backend=backend))
    assert result.value == 10



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_negative_int_positional(backend: Backend):
    @dataclass
    class Args:
        value: int

    result = parse(Args, "-5", backend=backend)
    assert result == Args(value=-5)


@backends
def test_negative_float_positional(backend: Backend):
    @dataclass
    class Args:
        value: float

    result = parse(Args, "-2.718", backend=backend)
    assert result == Args(value=-2.718)


@backends
def test_negative_positional_with_option(backend: Backend):
    @dataclass
    class Args:
        verbose: Annotated[bool, cappa.Arg(short=True, default=False)]
        value: int

    result = parse(Args, "-v", "-42", backend=backend)
    assert result == Args(verbose=True, value=-42)



from __future__ import annotations

from dataclasses import dataclass

import pytest
from typing_extensions import Annotated

import cappa
from cappa.parser import backend
from tests.utils import Backend, backends, parse


@backends
def test_invalid_choice_help(backend: Backend):
    @dataclass
    class Args:
        value: Annotated[str, cappa.Arg(long=True)]

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "--value", "ok", "--meow", "wat", backend=backend)
    assert e.value.code == 2
    assert "unrecognized arguments: --meow" in str(e.value.message).lower()


def test_invalid_choice_help_possible_options():
    @dataclass
    class Args:
        value: Annotated[str, cappa.Arg(long=True)]

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "--val", "on", backend=backend)
    assert e.value.code == 2
    assert "Unrecognized arguments: --val (Did you mean: --value)" in str(
        e.value.message
    )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Generic, TextIO, TypedDict

from typing_extensions import Unpack

import cappa
from cappa.base import Backend, CappaCapable, T
from cappa.help import HelpFormattable
from cappa.invoke.types import DepTypes
from cappa.state import State

__all__ = [
    "CommandRunner",
    "RunnerArgs",
]


class RunnerArgs(TypedDict, total=False):
    """Available kwargs for `parse` and `invoke` function, to match `CommandRunner` fields."""

    argv: list[str]
    backend: Backend | None
    output: cappa.Output | None
    color: bool
    version: str | cappa.Arg[str] | None
    help: bool | cappa.Arg[bool]
    completion: bool | cappa.Arg[bool]
    help_formatter: HelpFormattable | None
    input: TextIO | None
    state: State[Any] | None


@dataclass
class CommandRunner(Generic[T]):
    """Object to hold common parse/invoke invocation state, for testing.

    Accepts almost identical inputs to that of `parse`/`invoke`. The notable
    deviation is `argv`.

    Whereas the original functions accept argv as a single-argument list;
    `CommandRunner` accepts `base_args` at the class-constructor level, which
    is concatenated with the `CommandRunner.parse` or `CommandRunner.invoke`
    `*args`, to arrive a the total set of input args.

    Example:
        Some base CLI object

        >>> from dataclasses import dataclass
        >>>
        >>> import cappa
        >>>
        >>> @dataclass
        ... class Obj:
        ...     first: str
        ...     second: str = "2"

        Create an instance with no arguments means there is no default state

        >>> runner = CommandRunner()
        >>> runner.parse(Obj, "one")
        Obj(first='one', second='2')

        Or create a runner that always uses the same base CLI object, and default base command

        >>> runner = CommandRunner(Obj, base_args=["first"])

        Now each test, can customize the behavior to suit the test in question.

        >>> runner.parse(color=False)
        Obj(first='first', second='2')

        >>> runner.parse("two")
        Obj(first='first', second='two')
    """

    obj: CappaCapable[T] | None = None
    deps: DepTypes = None
    backend: Backend | None = None
    output: cappa.Output | None = None
    color: bool = True
    version: str | cappa.Arg[str] | None = None
    help: bool | cappa.Arg[bool] = True
    completion: bool | cappa.Arg[bool] = True
    help_formatter: HelpFormattable | None = None
    input: TextIO | None = None
    state: State[Any] | None = None

    base_args: list[str] = field(default_factory=lambda: [])

    def coalesce_kwargs(self, *args: str, **kwargs: Unpack[RunnerArgs]) -> RunnerArgs:
        return RunnerArgs(
            argv=self.base_args + list(args),
            backend=kwargs.get("backend") or self.backend,
            output=kwargs.get("output") or self.output,
            color=kwargs.get("color") or self.color,
            version=kwargs.get("version") or self.version,
            help=kwargs["help"] if "help" in kwargs else self.help,
            completion=kwargs["completion"]
            if "completion" in kwargs
            else self.completion,
            help_formatter=kwargs.get("help_formatter") or self.help_formatter,
            input=kwargs.get("input") or self.input,
            state=kwargs.get("state") or self.state,
        )

    def collect_args(
        self, obj: CappaCapable[T] | str | None = None, *args: str
    ) -> tuple[CappaCapable[T], tuple[str, ...]]:
        if isinstance(obj, str):
            assert self.obj
            return self.obj, (obj, *args)

        obj = obj or self.obj
        assert obj
        return obj, args

    def parse(
        self,
        obj: CappaCapable[T] | str | None = None,
        *args: str,
        **kwargs: Unpack[RunnerArgs],
    ) -> T:
        obj, args = self.collect_args(obj, *args)
        final_kwargs = self.coalesce_kwargs(*args, **kwargs)
        return cappa.parse(obj, **final_kwargs)

    def invoke(
        self,
        obj: CappaCapable[T] | str | None = None,
        *args: str,
        deps: DepTypes = None,
        **kwargs: Unpack[RunnerArgs],
    ) -> Any:
        obj, args = self.collect_args(obj, *args)
        final_kwargs = self.coalesce_kwargs(*args, **kwargs)
        deps = deps or self.deps
        return cappa.invoke(obj, **final_kwargs, deps=deps)

    async def invoke_async(
        self,
        obj: CappaCapable[T] | str | None = None,
        *args: str,
        deps: DepTypes = None,
        **kwargs: Unpack[RunnerArgs],
    ) -> Any:
        obj, args = self.collect_args(obj, *args)
        final_kwargs = self.coalesce_kwargs(*args, **kwargs)
        deps = deps or self.deps
        return await cappa.invoke_async(obj, **final_kwargs, deps=deps)



from __future__ import annotations

import sys
from dataclasses import dataclass
from typing import Literal

import pytest

from tests.utils import Backend, backends, parse


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires 3.10")
@backends
def test_valid(backend: Backend):
    @dataclass
    class ArgTest:
        name: Literal["one"] | Literal["two"] | Literal["three"] | Literal[4]

    test = parse(ArgTest, "two", backend=backend)
    assert test.name == "two"



from __future__ import annotations

import csv
import io
import logging
from dataclasses import dataclass
from typing import Any

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, invoke

log = logging.getLogger("test")


def binary_io():
    print("binary before")
    with io.BytesIO() as buffer:
        yield buffer
    print("binary after")


def text_io(binary_io: Annotated[io.BytesIO, cappa.Dep(binary_io)]):
    print("text before")
    text_buffer = io.TextIOWrapper(binary_io)
    yield text_buffer
    print("text after")


def csv_file(text_io: Annotated[io.TextIOWrapper, cappa.Dep(text_io)]):
    print("csv before")
    writer = csv.DictWriter(text_io, ["foo", "bar"])
    writer.writeheader()
    writer.writerow({"foo": 1, "bar": "woah"})
    text_io.detach()
    yield writer
    print("csv after")


def command(
    _: Annotated[dict[str, Any], cappa.Dep(csv_file)],
    binary_io: Annotated[io.BytesIO, cappa.Dep(binary_io)],
):
    assert not binary_io.closed
    return binary_io, binary_io.getvalue()


@cappa.command(invoke=command)
@dataclass
class Command: ...


@backends
def test_invoke_top_level_command(backend: Backend, capsys: Any):
    buffer, result = invoke(Command, backend=backend)

    assert buffer.closed is True
    assert result == b"foo,bar\r\n1,woah\r\n"

    out = capsys.readouterr().out
    assert out == (
        "binary before\ntext before\ncsv before\ncsv after\ntext after\nbinary after\n"
    )



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class Args:
    numbers: Annotated[str, cappa.Arg(parse=int, parse_inference=False)]


@backends
def test_disable_default_parser(backend: Backend):
    test = parse(Args, "1", backend=backend)
    assert test.numbers == 1



from __future__ import annotations

import asyncio
from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from cappa.default import ValueFrom
from tests.utils import Backend, backends


async def async_value_with_kwargs(multiplier: int) -> int:
    await asyncio.sleep(0.001)
    return 10 * multiplier


@dataclass
class Command:
    value: Annotated[
        int, cappa.Arg(default=ValueFrom(async_value_with_kwargs, multiplier=5))
    ]

    def __call__(self):
        return self


@backends
def test(backend: Backend):
    result = asyncio.run(cappa.invoke_async(Command, argv=[], backend=backend))
    assert result.value == 50



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class Foo: ...


@dataclass
class Args:
    subcommand: Annotated[Foo, cappa.Subcommand(group="Yup")]


@backends
def test_required_missing(backend: Backend, capsys: Any):
    with pytest.raises(cappa.Exit) as e:
        parse(Args, "--help", backend=backend)
    assert e.value.code == 0

    help = capsys.readouterr().out
    assert re.match(r".*Yup:?[\n\s]+\{?foo.*", help, re.DOTALL)


@backends
def test_explicit_group(backend: Backend, capsys: Any):
    @dataclass
    class Args:
        subcommand: Annotated[Foo, cappa.Subcommand(group=cappa.Group(name="Yup"))]

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "--help", backend=backend)
    assert e.value.code == 0

    help = capsys.readouterr().out
    assert re.match(r".*Yup:?[\n\s]+\{?foo.*", help, re.DOTALL)


@backends
def test_tuple_group(backend: Backend, capsys: Any):
    @dataclass
    class Args:
        subcommand: Annotated[Foo, cappa.Subcommand(group=(1, "Yup"))]

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "--help", backend=backend)
    assert e.value.code == 0

    help = capsys.readouterr().out
    assert re.match(r".*Yup:?[\n\s]+\{?foo.*", help, re.DOTALL)



from dataclasses import dataclass
from io import StringIO
from typing import Type

import pytest
from rich.prompt import Confirm as RichConfirm
from rich.prompt import Prompt as RichPrompt
from typing_extensions import Annotated

import cappa
from cappa import Confirm, Prompt
from cappa.base import Backend
from tests.utils import backends, parse

prompt_types = pytest.mark.parametrize("prompt_type", (Prompt, RichPrompt))
confirm_types = pytest.mark.parametrize("confirm_type", (Confirm, RichConfirm))


@backends
class TestPrompt:
    @prompt_types
    def test_value(self, backend: Backend, prompt_type: Type[Prompt]):
        @dataclass
        class Test:
            name: Annotated[str, cappa.Arg(default=prompt_type("gimme text number"))]

        result = parse(Test, backend=backend, input=StringIO("two"))

        assert result.name == "two"

    @prompt_types
    def test_default(self, backend: Backend, prompt_type: Type[Prompt]):
        @dataclass
        class Test:
            name: Annotated[
                str, cappa.Arg(default=prompt_type("gimme text number"))
            ] = "one"

        result = parse(Test, backend=backend, input=StringIO(""))

        assert result.name == "one"

    @prompt_types
    def test_mapped(self, backend: Backend, prompt_type: Type[Prompt]):
        @dataclass
        class Test:
            num: Annotated[int, cappa.Arg(default=prompt_type("gimme text number"))]

        result = parse(Test, backend=backend, input=StringIO("5"))

        assert result.num == 5


@backends
class TestConfirm:
    @confirm_types
    def test_value(self, backend: Backend, confirm_type: Type[Confirm]):
        @dataclass
        class Test:
            ok: Annotated[bool, cappa.Arg(default=confirm_type("gimme text number"))]

        result = parse(Test, backend=backend, input=StringIO("y"))

        assert result.ok is True

    @confirm_types
    def test_default(self, backend: Backend, confirm_type: Type[Confirm]):
        @dataclass
        class Test:
            ok: Annotated[
                bool, cappa.Arg(default=confirm_type("gimme text number"))
            ] = False

        result = parse(Test, backend=backend, input=StringIO(""))

        assert result.ok is False



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Union

import pytest

import cappa
from cappa.help import HelpFormatter, format_subcommand
from cappa.subcommand import FinalSubcommand
from tests.utils import Backend, backends, parse, parse_completion


@cappa.command(aliases=["ls"])
@dataclass
class List:
    pass


@cappa.command(
    aliases=[cappa.Alias("rm", deprecated="use remove")],
)
@dataclass
class Remove:
    pass


@dataclass
class Tool:
    cmd: cappa.Subcommands[Union[List, Remove]]


@backends
def test_visible_alias_dispatches(backend: Backend):
    result = parse(Tool, "list", backend=backend)
    assert isinstance(result.cmd, List)

    result = parse(Tool, "ls", backend=backend)
    assert isinstance(result.cmd, List)


@backends
def test_canonical_still_dispatches(backend: Backend):
    result = parse(Tool, "remove", backend=backend)
    assert isinstance(result.cmd, Remove)


@backends
def test_deprecated_alias_warns(backend: Backend, capsys: Any):
    result = parse(Tool, "rm", backend=backend)
    assert isinstance(result.cmd, Remove)

    err = capsys.readouterr().err
    assert "Command alias `rm` is deprecated: use remove" in err


@backends
def test_canonical_does_not_warn(backend: Backend, capsys: Any):
    parse(Tool, "remove", backend=backend)
    err = capsys.readouterr().err
    assert "deprecated" not in err


@cappa.command(aliases=[cappa.Alias("del", deprecated=True)])
@dataclass
class DeleteBool:
    pass


@dataclass
class DeleteBoolRoot:
    cmd: cappa.Subcommands[DeleteBool]


@backends
def test_deprecated_alias_bool_warns_without_suffix(backend: Backend, capsys: Any):
    """`deprecated=True` (bool) emits the default message with no `:` suffix."""
    result = parse(DeleteBoolRoot, "del", backend=backend)
    assert isinstance(result.cmd, DeleteBool)

    err = capsys.readouterr().err
    assert "Command alias `del` is deprecated" in err
    assert "Command alias `del` is deprecated:" not in err


# --- Hidden alias ---


@cappa.command(aliases=[cappa.Alias("dir", hidden=True)])
@dataclass
class HiddenList:
    pass


@dataclass
class HiddenTool:
    cmd: cappa.Subcommands[HiddenList]


@backends
def test_hidden_alias_dispatches(backend: Backend):
    result = parse(HiddenTool, "dir", backend=backend)
    assert isinstance(result.cmd, HiddenList)


@backends
def test_hidden_alias_not_in_help(backend: Backend, capsys: Any):
    with pytest.raises(cappa.Exit):
        parse(HiddenTool, "--help", backend=backend)

    out = capsys.readouterr().out
    assert "dir" not in out
    assert "hidden-list" in out


# --- Completion ---


def test_completion_includes_visible_alias():
    result = parse_completion(Tool, "")
    assert result is not None
    assert "list" in result
    assert "ls" in result
    assert "remove" in result
    assert "rm" in result


def test_completion_excludes_hidden_alias():
    result = parse_completion(HiddenTool, "")
    assert result is not None
    assert "hidden-list" in result
    assert "dir" not in result


# --- Help rendering ---


@backends
def test_visible_aliases_in_help(backend: Backend, capsys: Any):
    with pytest.raises(cappa.Exit):
        parse(Tool, "--help", backend=backend)

    out = capsys.readouterr().out
    assert "list, ls" in out
    assert "remove, rm (deprecated)" in out


# --- Imperative construction ---


@dataclass
class ImpA:
    pass


@dataclass
class ImpB:
    pass


@backends
def test_imperative_aliases(backend: Backend):
    """Aliases declared on Command directly (not via decorator) work the same way."""
    sub = cappa.Subcommand(
        field_name="cmd",
        options={
            "imp-a": cappa.Command(ImpA, name="imp-a", aliases=["a"]),
            "imp-b": cappa.Command(ImpB, name="imp-b", aliases=["b"]),
        },
    )

    @dataclass
    class ImpRoot:
        cmd: cappa.Subcommands[Union[ImpA, ImpB]]

    cmd = cappa.Command(ImpRoot, arguments=[sub])
    result = parse(cmd, "a", backend=backend)
    assert isinstance(result.cmd, ImpA)

    result = parse(cmd, "b", backend=backend)
    assert isinstance(result.cmd, ImpB)


# --- Collisions ---


@cappa.command(aliases=["self-collide"])
@dataclass
class SelfCollide:
    pass


@dataclass
class SelfCollideRoot:
    cmd: cappa.Subcommands[SelfCollide]


def test_alias_collides_with_own_canonical_name():
    with pytest.raises(ValueError, match="matching its own name"):
        cappa.collect(SelfCollideRoot)


@cappa.command(name="collide-x")
@dataclass
class CollideX:
    pass


@cappa.command(aliases=["collide-x"])
@dataclass
class CollideY:
    pass


@dataclass
class CollideXYRoot:
    cmd: cappa.Subcommands[Union[CollideX, CollideY]]


def test_alias_collides_with_other_canonical_name():
    with pytest.raises(ValueError, match="collides with another subcommand"):
        cappa.collect(CollideXYRoot)


@cappa.command(aliases=["shared"])
@dataclass
class CollideA:
    pass


@cappa.command(aliases=["shared"])
@dataclass
class CollideB:
    pass


@dataclass
class CollideABRoot:
    cmd: cappa.Subcommands[Union[CollideA, CollideB]]


def test_alias_collides_with_other_alias():
    with pytest.raises(ValueError, match="declared on both"):
        cappa.collect(CollideABRoot)


# --- Defensive helpers (direct unit calls) ---


def test_visible_aliases_for_unknown_canonical_returns_empty():
    """`visible_aliases_for` returns [] when given a name that isn't a subcommand."""
    sub = FinalSubcommand(field_name="cmd", options={})
    assert sub.visible_aliases_for("not-a-subcommand") == []


def test_format_subcommand_without_subcommand_arg():
    """`format_subcommand` works with the default `subcommand=None`."""
    cmd = cappa.Command(List).collect()
    padding, help_text = format_subcommand(HelpFormatter(), cmd)
    rendered = str(padding.renderable)
    assert "list" in rendered
    assert "ls" not in rendered  # aliases only appear when subcommand is provided
    assert help_text == cmd.help



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class RequiredMissingOne:
    foo: Annotated[Union[str, None], cappa.Arg(long=True)] = None


@dataclass
class RequiredMissing:
    subcommand: Annotated[RequiredMissingOne, cappa.Subcommand]


@backends
def test_required_missing(backend: Backend):
    with pytest.raises(cappa.Exit) as e:
        parse(RequiredMissing, backend=backend)
    assert isinstance(e.value.message, str)

    message = e.value.message.lower()
    if backend:
        assert "the following arguments are required: {required-missing-one}" in message
    else:
        assert (
            "a command is required: {[cappa.subcommand]required-missing-one[/cappa.subcommand]}"
            in message
        )


@dataclass
class RequiredProvidedOne:
    foo: Annotated[Union[str, None], cappa.Arg(long=True)] = None


@dataclass
class RequiredProvidedTwo:
    bar: Annotated[Union[str, None], cappa.Arg(long=True)] = None


@dataclass
class RequiredProvided:
    subcommand: Annotated[
        Union[RequiredProvidedOne, RequiredProvidedTwo], cappa.Subcommand()
    ]


@backends
def test_required_provided(backend: Backend):
    test = parse(
        RequiredProvided, "required-provided-one", "--foo", "foo", backend=backend
    )
    assert isinstance(test.subcommand, RequiredProvidedOne)
    assert test.subcommand.foo == "foo"

    test = parse(RequiredProvided, "required-provided-two", backend=backend)
    assert isinstance(test.subcommand, RequiredProvidedTwo)
    assert test.subcommand.bar is None

    test = parse(
        RequiredProvided, "required-provided-two", "--bar", "bar", backend=backend
    )
    assert isinstance(test.subcommand, RequiredProvidedTwo)
    assert test.subcommand.bar == "bar"


@cappa.command(name="one")
@dataclass
class NamedSubcommandOne:
    pass


@dataclass
class NamedSubcommand:
    subcommand: Annotated[NamedSubcommandOne, cappa.Subcommand()]


@backends
def test_named_subcommand(backend: Backend):
    test = parse(NamedSubcommand, "one", backend=backend)
    assert isinstance(test.subcommand, NamedSubcommandOne)



import sys
from dataclasses import dataclass
from typing import Any

import pytest
from rich.console import Console
from typing_extensions import Annotated

import cappa
from tests.utils import (
    Backend,
    backends,
    parse,
    strip_trailing_whitespace,
    terminal_width,
)


@pytest.mark.help
@backends
def test_arg_description_renders_markdown(backend: Backend, capsys: Any):
    @dataclass
    class Args:
        foo: str
        """`This` **is** _neat_"""

    with pytest.raises(cappa.Exit):
        parse(Args, "--help", backend=backend)

    result = capsys.readouterr()

    assert "This is neat" in result.out

    with pytest.raises(cappa.Exit):
        parse(
            Args,
            "--help",
            backend=backend,
            output=cappa.Output(
                output_console=Console(force_terminal=True, file=sys.stdout)
            ),
        )

    result = capsys.readouterr()
    assert "This\x1b[0m" in result.out
    assert " \x1b[1mis\x1b[0m" in result.out  # typos: ignore
    assert " \x1b[3mneat\x1b[0m" in result.out


@pytest.mark.help
@backends
def test_multi_paragraph_arg_help_preserves_paragraph_breaks(
    backend: Backend, capsys: Any
):
    @dataclass
    class Args:
        foo: Annotated[
            str,
            cappa.Arg(
                help="First paragraph summary.\n\nSecond paragraph body.",
            ),
        ]

    with terminal_width(80), pytest.raises(cappa.Exit):
        parse(Args, "--help", backend=backend)

    out = strip_trailing_whitespace(capsys.readouterr().out)

    assert "First paragraph summary." in out
    assert "Second paragraph body." in out

    summary_pos = out.index("First paragraph summary.")
    body_pos = out.index("Second paragraph body.")
    between = out[summary_pos + len("First paragraph summary.") : body_pos]
    assert "\n\n" in between


@pytest.mark.help
@backends
def test_soft_line_breaks_fold_into_spaces(backend: Backend, capsys: Any):
    @dataclass
    class SoftBreakArgs:
        foo: str
        """
        This is a long sentence that goes on and on.
        This second soft-break line continues the same paragraph.
        """

    with terminal_width(80), pytest.raises(cappa.Exit):
        parse(SoftBreakArgs, "--help", backend=backend, completion=False)

    out = strip_trailing_whitespace(capsys.readouterr().out)

    # Words from adjacent soft-break lines must be separated by a space
    assert "on.This" not in out
    assert "on. This" in out

    # And must not have a blank line between them (same paragraph)
    first = out.index("goes on and on.")
    second = out.index("same paragraph.")
    assert "\n\n" not in out[first:second]



import re
from dataclasses import dataclass
from textwrap import dedent
from typing import Any

import pytest

import cappa
from tests.utils import (
    Backend,
    backends,
    ignore_docstring_parser,
    parse,
    strip_trailing_whitespace,
)


@dataclass
class IncludesDocstring:
    """Does a thing.

    and does it really well!

    Args:
        foo (str): the value of foo
        bar (bool): whether to bar
    """

    foo: str
    bar: bool = False


@pytest.mark.help
@backends
def test_required_provided(backend: Backend, capsys: Any):
    with pytest.raises(cappa.Exit):
        parse(IncludesDocstring, "--help", backend=backend)

    result = capsys.readouterr().out

    assert "[--bar] FOO [-h]" in result
    assert re.match(r".*Does a thing\.\s+and does it really well!.*", result, re.DOTALL)
    assert re.match(r".*FOO\s+the value of foo.*", result, re.DOTALL)
    assert re.match(r".*\[--bar\]\s+whether to bar.*", result, re.DOTALL)


@pytest.mark.help
@backends
def test_just_a_title(backend: Backend, capsys: Any):
    @dataclass
    class IncludesDocstring:
        """Just a title."""

    with pytest.raises(cappa.Exit):
        parse(IncludesDocstring, "--help", backend=backend)

    result = capsys.readouterr().out

    assert "Just a title" in result


@pytest.mark.help
@backends
def test_docstring_with_explicit_help(backend: Backend, capsys: Any):
    @cappa.command(help="help text")
    @dataclass
    class IncludesDocstring:
        """Just a title."""

    with pytest.raises(cappa.Exit):
        parse(IncludesDocstring, "--help", backend=backend)

    result = capsys.readouterr().out

    assert "Just a title" not in result
    assert "help text" in result


@pytest.mark.help
@backends
def test_docstring_with_explicit_description(backend: Backend, capsys: Any):
    @cappa.command(description="description")
    @dataclass
    class IncludesDocstring:
        """Just a title."""

    with pytest.raises(cappa.Exit):
        parse(IncludesDocstring, "--help", backend=backend)

    result = capsys.readouterr().out

    assert "Just a title" in result
    assert "description" in result


@pytest.mark.help
@backends
def test_docstring_being_used_but_not_parsed(
    backend: Backend, capsys: Any, monkeypatch: Any
):
    @dataclass
    class UnparsedDocstring:
        """Summary.

        Example:
         - one
        """

    with pytest.raises(cappa.Exit), ignore_docstring_parser(monkeypatch):
        parse(UnparsedDocstring, "--help", backend=backend, completion=False)

    result = strip_trailing_whitespace(capsys.readouterr().out)

    assert result == dedent(
        """\
        Usage: unparsed-docstring [-h]

          Summary.
        
          Example:

           • one

          Help
            [-h, --help]  Show this message and exit.
        """
    )


@pytest.mark.help
@backends
def test_docstring_being_used_but_not_parsed_one_line(
    backend: Backend, capsys: Any, monkeypatch: Any
):
    @dataclass
    class UnparsedDocstring:
        """Summary."""

    with pytest.raises(cappa.Exit), ignore_docstring_parser(monkeypatch):
        parse(UnparsedDocstring, "--help", backend=backend, completion=False)

    result = strip_trailing_whitespace(capsys.readouterr().out)
    assert result == dedent(
        """\
        Usage: unparsed-docstring [-h]

          Summary.

          Help
            [-h, --help]  Show this message and exit.
        """
    )


@pytest.mark.help
@backends
def test_escaped_markdown(backend: Backend, capsys: Any, monkeypatch: Any):
    @dataclass
    class Escaped:
        """Blabla.

        And configure `~/.pypirc`:

        ```ini
        [distutils]
        index-servers =
            pypi
        ```
        """

    with pytest.raises(cappa.Exit), ignore_docstring_parser(monkeypatch):
        parse(Escaped, "--help", backend=backend, completion=False)

    result = capsys.readouterr().out
    for char in "┌─┐│└┘":
        result = result.replace(char, "")
    result = strip_trailing_whitespace(result)

    assert result == dedent(
        """\
        Usage: escaped [-h]

          Blabla.
          
          And configure ~/.pypirc:
          
          
           [distutils]
           index-servers =
               pypi


          Help
            [-h, --help]  Show this message and exit.
        """
    )


@pytest.mark.help
@backends
def test_attribute_docstring(backend: Backend, capsys: Any):
    @dataclass
    class Args:
        """Bah.

        Arguments:
            top_level: woo woo
            foo: this should get superseded
        """

        top_level: int

        foo: int
        """This is a foo."""

        bar: str
        """This is a bar."""

    with pytest.raises(cappa.Exit):
        parse(Args, "--help", backend=backend, completion=False)

    result = strip_trailing_whitespace(capsys.readouterr().out)
    assert result == dedent(
        """\
        Usage: args TOP_LEVEL FOO BAR [-h]

          Bah.

          Arguments
            TOP_LEVEL     woo woo
            FOO           This is a foo.
            BAR           This is a bar.

          Help
            [-h, --help]  Show this message and exit.
        """
    )


@pytest.mark.help
@backends
def test_explicit_help_description_manual_args(backend: Backend, capsys: Any):
    @cappa.command(help="Title", description="longer description")
    @dataclass
    class Args2:
        """...

        Args:
            foo: nope
            bar: yep
        """

        foo: int
        """this is a foo"""

        bar: int

    with pytest.raises(cappa.Exit):
        parse(Args2, "--help", backend=backend, completion=False)

    result = strip_trailing_whitespace(capsys.readouterr().out)
    assert result == dedent(
        """\
        Usage: args2 FOO BAR [-h]

          Title

          longer description

          Arguments
            FOO           this is a foo
            BAR           yep

          Help
            [-h, --help]  Show this message and exit.
        """
    )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_num_args_unbounded_length_num_args(backend: Backend):
    @dataclass
    class Args:
        a: Annotated[list[str], cappa.Arg(short=True, num_args=-1)] = field(
            default_factory=lambda: []
        )
        b: Annotated[str, cappa.Arg(short=True)] = field(default="")
        foo: Optional[str] = None

    # Unbounded length by itself
    t1 = parse(Args, backend=backend)
    assert t1 == Args(a=[], foo=None)

    # Unbounded length by itself
    t1 = parse(Args, "-a", "1", "2", "3", backend=backend)
    assert t1 == Args(a=["1", "2", "3"], foo=None)

    # An option can follow to terminate the option
    t1 = parse(Args, "-a", "1", "2", "3", "-b", "b", "foo", backend=backend)
    assert t1 == Args(a=["1", "2", "3"], b="b", foo="foo")

    # Or the -- separator can be used to terminate it
    t1 = parse(Args, "-a", "1", "2", "3", "--", "foo", backend=backend)
    assert t1 == Args(a=["1", "2", "3"], foo="foo")


@backends
def test_unbounded_positional_args(backend: Backend):
    @dataclass
    class Args:
        a: list[str]

    with pytest.raises(cappa.Exit) as e:
        parse(Args, backend=backend)
    error = str(e.value.message).lower()

    if backend:
        assert error == "the following arguments are required: a"
    else:
        assert error == "argument 'a' requires at least one values, found 0"

    t1 = parse(Args, "a", backend=backend)
    assert t1 == Args(["a"])



"""Tests for field_name grouping in help text."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, CapsysOutput, backends, parse


@backends
def test_basic_field_grouping(backend: Backend, capsys: Any):
    """Test that multiple args with same field_name are shown on same line."""

    @dataclass
    class Args:
        verbose: Annotated[
            bool,
            cappa.Arg(short="-v", long="--verbose", help="Enable verbose output."),
            cappa.Arg(long="--debug", help="Enable debug mode."),
        ]

    with pytest.raises(cappa.Exit):
        parse(Args, "--help", backend=backend)

    output = CapsysOutput.from_capsys(capsys)
    # Bool flags combined: [-v, --verbose, --debug]
    assert "[-v, --verbose, --debug]" in output.stdout
    # Help texts should be concatenated (may wrap across lines)
    assert "Enable verbose output." in output.stdout
    assert "debug mode." in output.stdout


@backends
def test_single_arg_unaffected(backend: Backend, capsys: Any):
    """Test that single args without field_name grouping are unaffected."""

    @dataclass
    class Args:
        name: Annotated[str, cappa.Arg(short="-n", help="Your name.")]

    with pytest.raises(cappa.Exit):
        parse(Args, "--help", backend=backend)

    output = CapsysOutput.from_capsys(capsys)
    assert "-n NAME" in output.stdout
    assert "Your name." in output.stdout


@backends
def test_multiple_grouped_fields(backend: Backend, capsys: Any):
    """Test multiple different fields with grouping."""

    @dataclass
    class Args:
        verbose: Annotated[
            int,
            cappa.Arg(
                short="-v", help="Verbosity level (lowercase v).", group="Verbose"
            ),
            cappa.Arg(
                short="-V", help="Verbosity level (uppercase V).", group="Verbose"
            ),
        ] = 0

        output: Annotated[
            str,
            cappa.Arg(short="-o", help="Output file.", group="Output"),
            cappa.Arg(short="-O", help="Alternative output option.", group="Output"),
        ] = "out.txt"

    with pytest.raises(cappa.Exit):
        parse(Args, "--help", backend=backend)

    output = CapsysOutput.from_capsys(capsys)

    # Check verbose group
    assert "[-v VERBOSE, -V VERBOSE]" in output.stdout
    assert "Verbosity level (lowercase v)." in output.stdout
    assert "(uppercase V)." in output.stdout

    # Check output group
    assert "[-o OUTPUT, -O OUTPUT]" in output.stdout
    assert "Output file." in output.stdout
    assert "Alternative" in output.stdout


@backends
def test_bool_pair_combined(backend: Backend, capsys: Any):
    """Test that --foo/--no-foo pairs are combined on the same line."""

    @dataclass
    class Args:
        debug: Annotated[bool, cappa.Arg(long="--debug/--no-debug")] = False

    with pytest.raises(cappa.Exit):
        parse(Args, "--help", backend=backend)

    output = CapsysOutput.from_capsys(capsys)

    # Bool flags with no value should be combined into single brackets
    assert "[--debug, --no-debug]" in output.stdout


@backends
def test_field_grouping_with_defaults(backend: Backend, capsys: Any):
    """Test that defaults are shown correctly for grouped fields."""

    @dataclass
    class Args:
        level: Annotated[
            int,
            cappa.Arg(short="-l", help="Level option 1."),
            cappa.Arg(short="-L", help="Level option 2."),
        ] = 5

    with pytest.raises(cappa.Exit):
        parse(Args, "--help", backend=backend)

    output = CapsysOutput.from_capsys(capsys)

    # Both args should be on the same line
    assert "[-l LEVEL, -L LEVEL]" in output.stdout
    # Default should appear
    assert "Default: 5" in output.stdout


@backends
def test_same_value_name_combined(backend: Backend, capsys: Any):
    """Test that args with same value_name show value once at end."""

    @dataclass
    class Args:
        verbose: Annotated[
            int,
            cappa.Arg(short="-v", help="Verbosity (short).", group="Options"),
            cappa.Arg(long="--verbose", help="Verbosity (long).", group="Options"),
        ] = 0

    with pytest.raises(cappa.Exit):
        parse(Args, "--help", backend=backend)

    output = CapsysOutput.from_capsys(capsys)

    # Format: [-v VERBOSE, --verbose VERBOSE]
    assert "[-v VERBOSE, --verbose VERBOSE]" in output.stdout


@backends
def test_different_value_names_each_shown(backend: Backend, capsys: Any):
    """Test that args with different value_names each show their own value."""

    @dataclass
    class Args:
        output: Annotated[
            str,
            cappa.Arg(
                short="-o", value_name="file", help="Output file.", group="Output"
            ),
            cappa.Arg(
                long="--output", value_name="path", help="Output path.", group="Output"
            ),
        ] = "out.txt"

    with pytest.raises(cappa.Exit):
        parse(Args, "--help", backend=backend)

    output = CapsysOutput.from_capsys(capsys)

    # Different value_names: [-o FILE, --output PATH]
    assert "[-o FILE, --output PATH]" in output.stdout


@backends
def test_mixed_value_names_with_none(backend: Backend, capsys: Any):
    """Test args where some consume values and some don't."""

    @dataclass
    class Args:
        # This creates exclusive args with different behaviors
        mode: Annotated[
            int | bool,
            cappa.Arg(short="-m", value_name="mode", help="Mode value.", group="Mode"),
            cappa.Arg(
                long="--default",
                action=cappa.ArgAction.store_true,
                help="Use default.",
                group="Mode",
            ),
        ] = 0

    with pytest.raises(cappa.Exit):
        parse(Args, "--help", backend=backend)

    output = CapsysOutput.from_capsys(capsys)

    # Format: [-m MODE, --default] (--default is store_true so no value)
    assert "[-m MODE, --default]" in output.stdout



from __future__ import annotations

from dataclasses import dataclass

import pytest
from typing_extensions import Annotated

import cappa
from cappa.subcommand import Subcommands
from tests.utils import parse


@dataclass
class SubSub:
    foo: int


@dataclass
class InvalidSubcommandSub:
    sub: Subcommands[SubSub]


@dataclass
class InvalidSubcommand:
    attrs: Annotated[InvalidSubcommandSub, cappa.Arg.destructured()]


def test_invalid_subcommand():
    with pytest.raises(ValueError) as e:
        parse(InvalidSubcommand)
    assert (
        str(e.value)
        == "Only `Arg` is supported in the context of a destructured argument"
    )



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_unrecognized_post_dash_arg(backend: Backend):
    @dataclass
    class Args:
        value: Annotated[str, cappa.Arg(long=True)]

    result = parse(Args, "--value=val", backend=backend)
    assert result == Args(value="val")


@backends
def test_value_contains_equal(backend: Backend):
    @dataclass
    class Args:
        value: Annotated[str, cappa.Arg(long=True)]

    result = parse(Args, "--value='var == val'", backend=backend)
    assert result == Args(value="'var == val'")



from __future__ import annotations

import textwrap
from dataclasses import dataclass
from typing import Any

import pytest

import cappa
from tests.utils import parse, strip_trailing_whitespace


@dataclass
class Sub:
    arg: str


@dataclass
class Args:
    command: cappa.Subcommands[Sub]


def test_required_arg_subcommand_help(capsys: Any):
    """Assert short help is emitted for the subcommand (on bad input) rather than the root help."""
    with pytest.raises(cappa.Exit) as e:
        parse(Args, "sub")

    assert e.value.code == 2
    out = strip_trailing_whitespace(capsys.readouterr().err)

    assert out == textwrap.dedent(
        """\
        Usage: args sub ARG [-h]

        Error: Option 'arg' requires an argument
        """
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, invoke


def has_default(default: bool = True):
    return {"default": default}


def command(
    default: Annotated[dict[str, Any], cappa.Dep(has_default)],
):
    return {"one": True, **default}


@cappa.command(invoke=command)
@dataclass
class Command: ...


@backends
def test_invoke_top_level_command(backend: Backend):
    result = invoke(Command, backend=backend)
    assert result == {"one": True, "default": True}



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any

import pytest

import cappa
from tests.utils import parse


def test_string_group(capsys: Any):
    @dataclass
    class Args: ...

    with pytest.raises(cappa.HelpExit) as e:
        parse(Args, "--help", backend=cappa.backend)

    assert e.value.code == 0

    out = capsys.readouterr().out

    options = re.findall(r"Valid\s+options:\s+generate,\s+complete", out, re.MULTILINE)
    assert len(options) == 1



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import pytest
from typing_extensions import Annotated

import cappa
from cappa.subcommand import Subcommands
from tests.utils import CapsysOutput, parse


@dataclass
class Command:
    foo: Annotated[int, cappa.Arg(long=True, propagate=True, help="Everywhere")] = 1
    sub: cappa.Subcommands[One | Two | None] = None


@dataclass
class One: ...


@dataclass
class Two: ...


def test_propagated_arg():
    result = parse(Command)
    assert result == Command(1, None)

    result = parse(Command, "--foo=4")
    assert result == Command(4, None)

    result = parse(Command, "one", "--foo=5")
    assert result == Command(5, One())

    result = parse(Command, "two", "--foo=6")
    assert result == Command(6, Two())


def test_propagate_incompatible_with_argparse():
    with pytest.raises(RuntimeError):
        parse(Command, backend=cappa.argparse.backend)


@dataclass
class Required:
    foo: Annotated[int, cappa.Arg(long=True, propagate=True)]
    sub: cappa.Subcommands[One | Two | None] = None


def test_required_propagated_arg():
    with pytest.raises(cappa.Exit):
        parse(Required)

    result = parse(Required, "--foo=4")
    assert result == Required(4, None)

    result = parse(Required, "one", "--foo=5")
    assert result == Required(5, One())

    result = parse(Required, "two", "--foo=6")
    assert result == Required(6, Two())


@dataclass
class ChildOverride:
    foo: Annotated[int, cappa.Arg(long=True, propagate=True)] = 1
    sub: cappa.Subcommands[Child | None] = None


@dataclass
class Child:
    foo: Annotated[int, cappa.Arg(long=True)] = 2


def test_child_override():
    result = parse(ChildOverride)
    assert result == ChildOverride(1, None)

    result = parse(ChildOverride, "--foo=4")
    assert result == ChildOverride(4, None)

    result = parse(ChildOverride, "--foo=4", "child")
    assert result == ChildOverride(4, Child(2))

    result = parse(ChildOverride, "child")
    assert result == ChildOverride(1, Child(2))

    result = parse(ChildOverride, "child", "--foo=5")
    assert result == ChildOverride(1, Child(5))

    result = parse(ChildOverride, "--foo=3", "child", "--foo=6")
    assert result == ChildOverride(3, Child(6))


def test_propagate_requires_option():
    @dataclass
    class ChildOverride:
        foo: Annotated[int, cappa.Arg(propagate=True)]

    with pytest.raises(ValueError) as e:
        parse(ChildOverride)
    assert (
        str(e.value)
        == "`Arg.propagate` requires a non-positional named option (`short` or `long`)."
    )


def test_help_contains_propagated_arg(capsys: Any):
    @dataclass
    class Command:
        foo: Annotated[int, cappa.Arg(long=True, propagate=True, help="Everywhere")] = 1
        bar: Annotated[int, cappa.Arg(long=True, help="Nowhere")] = 1
        sub: cappa.Subcommands[One | Two | None] = None

    with pytest.raises(cappa.Exit):
        parse(Command, "--help")
    output = CapsysOutput.from_capsys(capsys)
    assert "foo" in output.stdout
    assert "Everywhere" in output.stdout
    assert "Nowhere" in output.stdout

    with pytest.raises(cappa.Exit):
        parse(Command, "one", "--help")
    output = CapsysOutput.from_capsys(capsys)
    assert "foo" in output.stdout
    assert "Everywhere" in output.stdout
    assert "Nowhere" not in output.stdout

    with pytest.raises(cappa.Exit):
        parse(Command, "two", "--help")
    output = CapsysOutput.from_capsys(capsys)
    assert "foo" in output.stdout
    assert "Everywhere" in output.stdout
    assert "Nowhere" not in output.stdout


def test_count():
    """Test that actions requiring accumulated prior state (e.g. count) **get** values from the correct context."""

    @dataclass
    class Command:
        foo: Annotated[int, cappa.Arg(short=True, propagate=True, count=True)] = 1
        subcommand: Subcommands[One | None] = None

    result = parse(Command, "-f")
    assert result == Command(1, None)

    result = parse(Command, "-fff")
    assert result == Command(3, None)

    result = parse(Command, "one", "-f")
    assert result == Command(1, One())

    result = parse(Command, "one", "-fff")
    assert result == Command(3, One())



from __future__ import annotations

import enum
from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import parse


class Flavor(enum.Enum):
    vanilla = "vanilla"
    chocolate = "chocolate"


@dataclass
class Attributes:
    color: Annotated[str, cappa.Arg(long="attribute.color")]
    flavors: Annotated[list[Flavor], cappa.Arg(value_name="attribute.flavor")]
    something: Annotated[int, cappa.Arg(short="-p")]
    custom_parse: Annotated[str, cappa.Arg(long="foo", parse=lambda _: "always this")]  # pyright: ignore
    custom_action: Annotated[int, cappa.Arg(long="bar", action=lambda: 42)]  # pyright: ignore
    optional: Annotated[int, cappa.Arg(long="baz")] = 4


@dataclass
class Args:
    attrs: Annotated[cappa.Destructured[Attributes], cappa.Arg(hidden=True)]


def test_destructured():
    test = parse(
        Args,
        "--attribute.color=red",
        "chocolate",
        "vanilla",
        "-p4",
        "--foo=whatever",
        "--bar=0",
    )
    assert test == Args(
        attrs=Attributes(
            "red",
            [Flavor.chocolate, Flavor.vanilla],
            something=4,
            custom_parse="always this",
            custom_action=42,
            optional=4,
        )
    )



from __future__ import annotations

import sys
from dataclasses import dataclass
from typing import IO, Any, BinaryIO, TextIO

import cappa


@dataclass
class FileMode:
    """Factory for creating file object types.

    Instances of FileType are typically passed as type= arguments to the
    ArgumentParser add_argument() method.

    Arguments:
        mode: The file mode to use to open the file. Passes directly through to builtin `open()`.
        buffering: The file's desired buffer size. Passes directly through to builtin `open()`.
        encoding: The file's encoding. Passes directly through to builtin `open()`.
        errors: A string indicating how encoding and decoding errors are to
            be handled. Passes directly through to builtin `open()`.

        error_code: The exit code to use when an error occurs. Defaults to 1. Note this is **not**
            an `open()` argument.
    """

    mode: str = "r"
    buffering: int = -1
    encoding: str | None = None
    errors: str | None = None

    error_code: int = 1

    def __call__(self, filename: str) -> IO[Any] | TextIO | BinaryIO:
        """Open the given `filename` and return the file handle.

        Supply "-" as the filename to read from stdin or write to stdout,
        depending on the chosen `mode`.
        """
        # the special argument "-" means sys.std{in,out}
        if filename == "-":
            if "r" in self.mode:
                if "b" in self.mode:
                    return sys.stdin.buffer
                return sys.stdin

            if any(c for c in self.mode if c in {"w", "a", "x"}):
                if "b" in self.mode:
                    return sys.stdout.buffer
                return sys.stdout

            raise cappa.Exit(
                f"Invalid mode '{self.mode}' with supplied '-' file name.",
                code=self.error_code,
            )

        try:
            return open(filename, self.mode, self.buffering, self.encoding, self.errors)
        except OSError as e:
            raise cappa.Exit(f"Cannot open {filename}: {e}", code=self.error_code)



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

from typing_extensions import Annotated

import cappa
from cappa import Subcommands
from tests.utils import Backend, backends, parse


@dataclass
class Example:
    subcommand: Subcommands[Union[A, None]] = None


@cappa.command(name="a")
@dataclass
class A:
    a: Annotated[str, cappa.Arg(value_name="<Str a>", short="-a")]


@backends
def test_value_name_uses_correct_value(backend: Backend):
    result = parse(Example, "a", "-a", "test", backend=backend)
    assert result == Example(subcommand=A(a="test"))



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

from typing_extensions import Annotated, Literal

import cappa
from tests.utils import parse_completion


def test_choices_without_help():
    @dataclass
    class Args:
        value: Union[Literal["one"], Literal["two"], Literal["three"]]

    result = parse_completion(Args, "")
    assert result
    assert result == "one:\ntwo:\nthree:"


def test_choices_with_help():
    @dataclass
    class Args:
        value: Annotated[
            Union[Literal["one"], Literal["two"], Literal["three"]],
            cappa.Arg(help="helpful"),
        ]

    result = parse_completion(Args, "")
    assert result
    assert result == "one:helpful\ntwo:helpful\nthree:helpful"


def test_half_string():
    @dataclass
    class Args:
        value: Union[Literal["one"], Literal["two"], Literal["three"]]

    result = parse_completion(Args, "'on")
    assert result
    assert result == "one:"



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import parse


@dataclass
class Config:
    color: Annotated[str, cappa.Arg(long=True)]


@dataclass
class Args:
    config: Annotated[Config, cappa.Arg.destructured()]


def test_destructured():
    test = parse(Args, "--color=red")
    assert test == Args(config=Config(color="red"))


@dataclass
class Args2:
    config: Annotated[Config, cappa.Arg(destructure=True)]


def test_destructured_alt():
    test = parse(Args2, "--color=red")
    assert test == Args2(config=Config(color="red"))



from __future__ import annotations

from dataclasses import dataclass

import pytest

import cappa
from tests.utils import Backend, backends, invoke


@backends
def test_invalid_dependency(backend: Backend):
    def command(levels: int):
        return levels

    @cappa.command(invoke=command)
    @dataclass
    class Command: ...

    with pytest.raises(RuntimeError) as e:
        invoke(Command, backend=backend)

    exc = e.value
    cause = exc.__cause__
    assert "due to resolution failure" in str(exc)
    assert "`levels: int` is not a valid dependency for Dep(command)." == str(cause)


@backends
def test_unannotated_argument(backend: Backend):
    def command(levels):  # pyright: ignore
        return levels  # pyright: ignore

    @cappa.command(invoke=command)  # pyright: ignore
    @dataclass
    class Command: ...

    with pytest.raises(RuntimeError) as e:
        invoke(Command, backend=backend)

    exc = e.value
    cause = exc.__cause__
    assert "due to resolution failure" in str(exc)
    assert "`levels: <empty>` is not a valid dependency for Dep(command)." == str(cause)



from rich.theme import Theme

default_theme: Theme = Theme(
    {
        "cappa.prog": "grey50",
        "cappa.group": "dark_orange bold",
        "cappa.arg": "cyan",
        "cappa.arg.name": "dark_cyan",
        "cappa.subcommand": "dark_cyan",
        "cappa.help": "",
    }
)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, invoke


def foo(command: cappa.Self[HasDefault]):
    print(f"called with {command.default}")


@dataclass
class HasDefault:
    default: Annotated[int, cappa.Arg(short=True)] = 1

    def __call__(self):
        return 2


@dataclass
class Command(HasDefault):
    sub: cappa.Subcommands[Foo | Bar | None] = None


@dataclass
class Foo(HasDefault):
    default: Annotated[int, cappa.Arg(short=True)] = 2


@dataclass
class Bar(HasDefault):
    default: Annotated[int, cappa.Arg(short=True)] = 3


@backends
def test_typing_self(capsys: Any, backend: Backend):
    invoke(Command, deps=[foo], backend=backend)
    out = capsys.readouterr().out
    assert "called with 1" in out

    invoke(Command, "foo", deps=[foo], backend=backend)
    out = capsys.readouterr().out
    assert "called with 2" in out

    invoke(Command, "bar", deps=[foo], backend=backend)
    out = capsys.readouterr().out
    assert "called with 3" in out



from __future__ import annotations

from dataclasses import dataclass
from datetime import date, datetime, time

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_datetime(backend: Backend):
    @dataclass
    class ArgTest:
        datetime_arg: datetime
        datetime_opt: Annotated[datetime, cappa.Arg(long=True)]

    test = parse(
        ArgTest,
        "2020-01-01T01:02:03",
        "--datetime-opt=2021-02-03T04:05:06",
        backend=backend,
    )
    assert test.datetime_arg == datetime(2020, 1, 1, 1, 2, 3)
    assert test.datetime_opt == datetime(2021, 2, 3, 4, 5, 6)


@backends
def test_date(backend: Backend):
    @dataclass
    class ArgTest:
        date_arg: date
        date_opt: Annotated[date, cappa.Arg(long=True)]

    test = parse(ArgTest, "2020-01-01", "--date-opt=2021-02-03", backend=backend)
    assert test.date_arg == date(2020, 1, 1)
    assert test.date_opt == date(2021, 2, 3)


@backends
def test_time(backend: Backend):
    @dataclass
    class ArgTest:
        date_arg: time
        date_opt: Annotated[time, cappa.Arg(long=True)]

    test = parse(ArgTest, "01:02:03", "--date-opt=04:05:06", backend=backend)
    assert test.date_arg == time(1, 2, 3)
    assert test.date_opt == time(4, 5, 6)



from __future__ import annotations

import pytest
from typing_extensions import Annotated

import cappa
from cappa.output import Exit, rich_to_ansi
from tests.utils import Backend, backends, parse


@cappa.command
class Args:
    flag: Annotated[
        str,
        cappa.Arg(long="--blue", action=lambda: "blue", num_args=0),
        cappa.Arg(long="--red", action=lambda: "red", num_args=0),
        cappa.Arg(long="--green", action=lambda: "green", num_args=0),
    ] = "blue"


@backends
def test_help(backend: Backend):
    output = cappa.Output()
    with pytest.raises(Exit) as e:
        parse(Args, "--help", backend=backend)

    assert "[--blue, --red, --green]" in rich_to_ansi(
        output.output_console, e.value.message
    )

    result = parse(Args, "--blue", backend=backend)
    assert result.flag == "blue"

    result = parse(Args, "--red", backend=backend)
    assert result.flag == "red"

    with pytest.raises(Exit) as e:
        parse(Args, "--blue", "--red", backend=backend)



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import parse


@dataclass
class Args:
    foo: Annotated[
        bool,
        cappa.Arg(
            long=["--foo", "--no-foo"],
            propagate=True,
            default=False,
            help="Enable or disable foo feature",
        ),
    ]
    subcommand: cappa.Subcommands[Sub]


@dataclass
class Sub:
    name: str


def test_propagated_arg():
    result = parse(Args, "--foo", "sub", "bar")
    assert result == Args(foo=True, subcommand=Sub(name="bar"))

    result = parse(Args, "--no-foo", "sub", "bar")
    assert result == Args(foo=False, subcommand=Sub(name="bar"))

    result = parse(Args, "sub", "--foo", "bar")
    assert result == Args(foo=True, subcommand=Sub(name="bar"))

    result = parse(Args, "sub", "--no-foo", "bar")
    assert result == Args(foo=False, subcommand=Sub(name="bar"))



from __future__ import annotations

import textwrap
from dataclasses import dataclass
from typing import Any

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import parse, strip_trailing_whitespace


def test_argument_name(capsys: Any):
    @dataclass
    class Args:
        name: Annotated[str, cappa.Arg(value_name="sname", help="more")]
        short: Annotated[str, cappa.Arg(short=True, value_name="ostr")]
        unbounded: Annotated[list[list[str]], cappa.Arg(short=True, value_name="UNB")]
        unbounded_pos: Annotated[list[str], cappa.Arg(value_name="upos", help="lots")]

    with pytest.raises(cappa.HelpExit) as e:
        parse(Args, "--help")

    assert e.value.code == 0

    out = strip_trailing_whitespace(capsys.readouterr().out)

    assert out == textwrap.dedent(
        """\
        Usage: args -s OSTR -u UNB ... SNAME UPOS ... [-h] [--completion COMPLETION]

          Options
            -s OSTR
            -u UNB ...

          Arguments
            SNAME                      more
            UPOS ...                   lots

          Help
            [-h, --help]               Show this message and exit.
            [--completion COMPLETION]  Use --completion generate to print shell-specific
                                       completion source. Valid options: generate,
                                       complete.
        """
    )



from __future__ import annotations

import enum
from dataclasses import dataclass

import pytest

import cappa
from tests.utils import Backend, backends, parse


class Options(enum.Enum):
    one = "one"
    two = "two"
    three = "three"


@dataclass
class ArgTest:
    options: Options


@backends
def test_valid(backend: Backend):
    test = parse(ArgTest, "two", backend=backend)
    assert test.options is Options.two


@backends
def test_invalid(backend: Backend):
    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "thename", backend=backend)

    message = str(e.value.message).lower()
    assert "invalid choice: 'thename' (choose from 'one', 'two', 'three')" in message
